# Licensed under GPL version 3 - see LICENSE.rst

import numpy as np
from astropy.units import Quantity
from astropy.table import Table
import astropy.units as u

from transforms3d.euler import euler2mat
from transforms3d.quaternions import mat2quat
from transforms3d.axangles import axangle2mat

from ...source import FixedPointing
from .fitsheaders import complete_header

class LissajousDither(FixedPointing):
    '''Lissajous dither pattern with adjustable parameters.

    Parameters
    ----------
    DitherAmp : `astropy.units.Quantity`
        (pitch, yaw, roll) dither amplitude
    DitherPeriod : `astropy.units.Quantity`
        (pitch, yaw, roll) dither Period in sec
    DitherPhase : `astropy.units.Quantity`
        (pitch, yaw, roll) dither phase at ``time = 0`` in radian
    '''
    def __init__(self, **kwargs):
        self.DitherAmp = kwargs.pop('DitherAmp', np.array([8., 8., 0.]) * u.arcsec)
        self.DitherPeriod = kwargs.pop('DitherPeriod', np.array([1000., 707., 1e5]) * u.s)
        self.DitherPhase = kwargs.pop('DitherPhase', np.zeros(3) * u.radian)
        super(LissajousDither, self).__init__(**kwargs)

    def dither(self, time):
        '''Calculate the dither offset relative to pointing direction.

        Parameters
        ----------
        time : np.array
            Time when the dither motion should be calculated

        Returns
        -------
        delta : np.array of shape (N, 3)
            dither motion offset in pitch, yaw, roll for N times in rad
        '''
        return self.DitherAmp * np.sin(2. * np.pi * u.radian * time[:, np.newaxis] * u.s / self.DitherPeriod + self.DitherPhase)  / np.array([np.cos(self.coords.icrs.dec), 1., 1.])


    def pointing(self, time):
        '''Calculate the pointing direction for a set of times
        There are the steps to convert the dither (which are offsets) to the total
        pointing direction with the roll properly taken care of.
        (See marxasp.c in the marx code for more details.)
        1.  Convert ra/dec offsets to absolute pointing.
        2.  Roll resulting vector about nominal pointing
        3.  Convert result to ra/dec.

        Parameters
        ----------
        time : np.array
            Array of times

        Returns
        -------
        pointing : (n, 3) np.array
            Ra, Dec, roll values in radian for the pointing direction at time t.
        '''
        nominal = Quantity([self.coords.ra, self.coords.dec,
                            self.roll])
        dither = self.dither(time)
        # roll in astronomical system is defined opposite of the usual mathematical angle
        # because the ra in the coordinate system increases in the other direction.
        roll = nominal[2] + dither[:, 2]
        # Express directions as x,y,z vectors
        phi = nominal[0]
        theta = np.pi/2. * u.rad - nominal[1]
        e_nominal = np.array([np.sin(phi) * np.sin(theta),
                              np.cos(phi) * np.sin(theta),
                              np.cos(theta)])
        phi = nominal[0] + dither[:, 0]
        theta = np.pi/2. * u.rad - (nominal[1] + dither[:, 1])
        e_dither = np.vstack([np.sin(phi) * np.sin(theta),
                              np.cos(phi) * np.sin(theta),
                              np.cos(theta)]).T
        pointing_dir = np.zeros_like(dither.value)

        # common case for Chandra
        if np.allclose(roll.to(u.rad).value, roll[0].to(u.rad).value):
            mat = axangle2mat(e_nominal, -roll[0].to(u.rad).value, is_normalized=True)
            constant_roll = True

        for i in range(len(time)):
            if not constant_roll:
                mat = axangle2mat(e_nominal, -roll[i].to(u.rad).value, is_normalized=True)
            pointing_dir[i, :] = np.dot(mat, e_dither[i, :].value)

        # convert x,y,z pointing back to ra, dec, roll
        pointing = np.vstack([np.arctan2(pointing_dir[:, 0], pointing_dir[:, 1]) % (2.*np.pi),
                              np.pi / 2. - np.arccos(pointing_dir[:, 2]),
                              roll.to(u.rad).value]).T
        return pointing

    def photons_dir(self, coos, time):
        '''Calculate direction on photons in homogeneous coordinates.

        Parameters
        ----------
        coos : `astropy.coordiantes.SkyCoord`
            Origin of each photon on the sky
        time : np.array
            Time for each photons in sec

        Returns
        -------
        photons_dir : np.array of shape (n, 4)
            Homogeneous direction vector for each photon
        '''
        ra = coos.ra.rad
        dec = coos.dec.rad
        # Minus sign here because photons start at +inf and move towards origin
        pointing = self.pointing(time)
        photons_dir = np.zeros((len(ra), 4))
        photons_dir[:, 0] = - np.cos(dec) * np.cos(ra)
        photons_dir[:, 1] = - np.cos(dec) * np.sin(ra)
        photons_dir[:, 2] = - np.sin(dec)
        for i in range(len(ra)):
            mat3d = euler2mat(pointing[i, 0],
                              - pointing[i, 1],
                              - pointing[i, 2], 'rzyx')

            photons_dir[i, :3] = np.dot(mat3d.T, photons_dir[i, :3])

        return photons_dir


    def write_asol(self, photons, asolfile, timestep=0.256):
        '''Write an aspect solution (asol) file

        Chandra analysis scripts often require an aspect solution, which is essientiall
        a list of pointing directions in the dither pattern vs. time.
        This method write such a list to a file.

        Parameters
        ----------
        photons :  `astropy.table.Table` or `astropy.table.Row`
            Table with photon properties. Some meta data from the header of this table
            is required (e.g. the length of the observation).
        asolfile : string
            Path and file name where the asol file is saved.
        timestamp : float
            Time step between entries in the asol file in seconds.
        '''
        time = np.arange(0, photons.meta['EXPOSURE'][0], timestep)
        pointing = self.pointing(time).to(u.deg).value
        asol = Table([time, pointing[:, 0], pointing[:, 1], pointing[:, 2]],
                     names=['time', 'ra', 'dec', 'roll'],
                     )
        asol['time'].unit = 's'
        # The following columns represent measured offsets in Chandra
        # They are not part of this simulation. Simply set them to 0
        for col in [ 'ra_err', 'dec_err', 'roll_err',
                     'dy', 'dz', 'dtheta', 'dy_err', 'dz_err', 'dtheta_err',
                      'roll_bias', 'pitch_bias', 'yaw_bias', 'roll_bias_err', 'pitch_bias_err', 'yaw_bias_err']:
            asol[col] = np.zeros_like(time)
            if 'bias' in col:
                asol[col].unit = 'deg / s'
            elif ('dy' in col) or ('dz' in col):
                asol[col].unit = 'mm'
            else:
                asol[col].unit = 'deg'
        asol['q_att'] = [mat2quat(euler2mat(np.deg2rad(p[0]),
                                   np.deg2rad(-p[1]),
                                   np.deg2rad(-p[2]), 'rzyx'))
                         for p in pointing]
        # Copy info like the exposure time from the photons list meta to asol,
        # but not column specific keywords like TTYPEn, TCTYPn, MTYPEn, MFORMn, etc:
        for k in photons.meta:
            if (('TYP' not in k) and ('FORM' not in k) and
                ('TCR' not in k) and ('TCDEL' not in k) and (k not in asol.meta)):
                asol.meta[k] = photons.meta[k]
        asol.meta['EXTNAME'] = 'ASPECT'
        complete_header(asol.meta, None, 'ACASOL', ['OGIP', 'TEMPORALDATA', 'ASPECT'])
        # In MARXS t=0 is the start of the observation, but for Chandra we need to make that
        # consistent with the value of the TSTART keyword.
        asol['time'] += asol.meta['TSTART'][0]
        asol.write(asolfile, format='fits')  # , checksum=True) - works only with fits interface



    # def pointing(self, time):
    #     '''Calculate the pointing direction for a set of times

    #     There are the steps to convert the dither (which are offsets) to the total
    #     pointing direction with the roll properly taken care of.
    #     (See marxasp.c in the marx code for more details.)

    #     1.  Convert ra/dec offsets to absolute pointing.
    #     2.  Roll resulting vector about nominal pointing
    #     3.  Convert result to ra/dec.

    #     Parameters
    #     ----------
    #     time : np.array
    #         Array of times

    #     Results
    #     -------
    #     pointing : (n, 3) np.array
    #         Ra, Dec, roll values in radian for the pointing direction at time t.
    #     '''
    #     dither = self.dither(time).to(u.radian).value
    #     phi = dither[:, 0]
    #     theta = np.pi/2. - dither[:, 1]
    #     e_dither = np.vstack([np.sin(phi) * np.sin(theta),
    #                           np.cos(phi) * np.sin(theta),
    #                           np.cos(theta)]).T
    #     pointing_dir = np.zeros_like(dither)

    #     # common case for Chandra
    #     roll = self.roll.to(u.radian).value
    #     if np.allclose(dither[:, 2], dither[0, 2]):
    #         mat = axangle2mat([1, 0, 0], -dither[0, 2] - roll, is_normalized=True)
    #         constant_roll = True

    #     for i in range(len(time)):
    #         if not constant_roll:
    #             mat = axangle2mat([1, 0, 0], -dither[i, 2] - roll, is_normalized=True)
    #         pointing_dir[i, :] = np.dot(mat, e_dither[i, :])

    #     # convert x,y,z pointing back to ra, dec, roll
    #     pointing = np.vstack([np.arctan2(pointing_dir[:, 0], pointing_dir[:, 1]) % (2.*np.pi),
    #                           np.pi / 2. - np.arccos(pointing_dir[:, 2]),
    #                           dither[:, 2]]).T
    #     return pointing

    # def photons_dir(self, coos, time):
    #     '''Calculate direction on photons in homogeneous coordinates.

    #     Parameters
    #     ----------
    #     coos : `astropy.coordiantes.SkyCoord`
    #         Origin of each photon on the sky
    #     time : np.array
    #         Time for each photons in sec

    #     Returns
    #     -------
    #     photons_dir : np.array of shape (n, 4)
    #         Homogeneous direction vector for each photon
    #     '''
    #     photons_dir = super(LissajousDither, self).photons_dir(coos, time)
    #     # Minus sign here because photons start at +inf and move towards origin
    #     pointing = self.dither(time).to(u.rad).value
    #     for i in range(len(time)):
    #         mat3d = euler2mat(pointing[i, 0],
    #                           - pointing[i, 1],
    #                           - pointing[i, 2], 'rzyx')

    #         photons_dir[i, :3] = np.dot(mat3d.T, photons_dir[i, :3])

    #     return photons_dir


    # def write_asol(self, photons, asolfile, timestep=0.256):
    #     time = np.arange(0, photons.meta['EXPOSURE'][0], timestep)
    #     pointing = np.rad2deg(self.pointing(time))
    #     asol = Table([time, pointing[:, 0], pointing[:, 1], pointing[:, 2]],
    #                  names=['time', 'ra', 'dec', 'roll'],
    #                  )
    #     asol['time'].unit = 's'
    #     # The following columns represent measured offsets in Chandra
    #     # They are not part of this simulation. Simply set them to 0
    #     for col in [ 'ra_err', 'dec_err', 'roll_err',
    #                  'dy', 'dz', 'dtheta', 'dy_err', 'dz_err', 'dtheta_err',
    #                   'roll_bias', 'pitch_bias', 'yaw_bias', 'roll_bias_err', 'pitch_bias_err', 'yaw_bias_err']:
    #         asol[col] = np.zeros_like(time)
    #         if 'bias' in col:
    #             asol[col].unit = 'deg / s'
    #         elif ('dy' in col) or ('dz' in col):
    #             asol[col].unit = 'mm'
    #         else:
    #             asol[col].unit = 'deg'
    #     asol['q_att'] = [mat2quat(euler2mat(np.deg2rad(p[0]),
    #                                np.deg2rad(-p[1]),
    #                                np.deg2rad(-p[2]), 'rzyx'))
    #                      for p in pointing]
    #     # Copy info like the exposure time from the photons list meta to asol,
    #     # but not column specific keywords like TTYPEn, TCTYPn, MTYPEn, MFORMn, etc:
    #     for k in photons.meta:
    #         if (('TYP' not in k) and ('FORM' not in k) and
    #             ('TCR' not in k) and ('TCDEL' not in k) and (k not in asol.meta)):
    #             asol.meta[k] = photons.meta[k]
    #     asol.meta['EXTNAME'] = 'ASPECT'
    #     complete_header(asol.meta, None, 'ACASOL', ['OGIP', 'TEMPORALDATA', 'ASPECT'])
    #     # In MARXS t=0 is the start of the observation, but for Chandra we need to make that
    #     # consistent with the value of the TSTART keyword.
    #     asol['time'] += asol.meta['TSTART'][0]
    #     asol.write(asolfile, format='fits')  # , checksum=True) - works only with fits interface



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest
from transforms3d import axangles

from ..rotations import ex2vec_fix, axangle2mat

def is_orthogonal(a):
    '''Return True is a matrix is orthonormal'''
    try:
        if (a.ndim == 2) and (a.shape[0] == a.shape[1]):
            return np.allclose(np.dot(a, a.T), np.eye(a.shape[0]))
        else:
            return False
    # e.g. `a` is not a matirx
    except:
        return False

def is_specialorthogonal(a):
    return is_orthogonal(a) and np.allclose(np.linalg.det(a), 1)

def test_ex2vec_fix():
    '''Valid inputs. All results have to be rotation matrices.'''
    assert np.all(ex2vec_fix(np.array([1., 0, 0]), np.array([0., 1, 0])) == np.eye(3))

    for i in [1., 1.2, 3.4, 5.6]:
        t = i + np.arange(3)
        rot = ex2vec_fix(t, np.array([0., 1. ,0.]))
        assert is_specialorthogonal(rot)

def test_ex2vec_fix_invalid_input():
    '''Rotation is not well defined, if both input vectors are parallel.'''
    with pytest.raises(ValueError) as e:
        v = np.array([1.2, 3.4, 5.6])
        t = ex2vec_fix(v, 3 * v)
    assert 'are parallel' in str(e.value)


def test_axangle2mat():
    '''Check that vectorized version gives same answers.'''
    axis = np.random.rand(4, 3)
    angles = np.arange(4.)
    out = axangle2mat(axis, angles)
    assert np.all(out[0, :, :] == np.eye(3))
    for i in range(3):
        out1 = axangles.axangle2mat(axis[i + 1, :], angles[i + 1])
        assert np.allclose(out[i + 1, :, :], out1)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

from ..random import RandomArbitraryPdf

# Any number will do. Just make it repeatable.
np.random.seed(12324)

def test_return_right_pdf():
    x = np.arange(10.)
    f = x ** 2.
    rand = RandomArbitraryPdf(x, f)
    draws = rand(10000)
    draws.sort()
    # there should be very few small numbers
    assert draws[5000] > 5
    # but a few of them
    assert draws[2] <= 3
    # check range
    assert draws[0] >=0
    assert draws[-1] < 10.

def test_unequal_bins():
    '''Almost all results should be from the 1-100 range because we are talking
    probability DENSITY.'''
    rand = RandomArbitraryPdf(np.array([0,1,100]), np.ones(3))
    draws = rand(10000)
    draws.sort()
    assert draws[1000] > 1



# Licensed under GPL version 3 - see LICENSE.rst
from copy import copy, deepcopy

import numpy as np
from transforms3d.affines import decompose44, compose
from transforms3d.axangles import axangle2mat

from .utils import e2h, h2e, anglediff, angle_between, norm_vector
from .pluecker import point_dir2plane
from ..base import _parse_position_keywords
from ..visualization.utils import plane_with_hole, combine_disjoint_triangulations

__all__ = ['xyz_square', 'xyz_circle',
           'NoGeometry', 'Geometry',
           'FinitePlane', 'PlaneWithHole',
           'RectangleHole', 'CircularHole',
           'Cylinder',
           ]


def xyz_square(geometry, r_factor=1):
    '''Generate Eukledian positions for the corners of a square.

    The square is centered on the center of the object and the edges are
    given by ``v_y`` and ``v_z``.

    Parameters
    ----------
    r_factor : float
        Scaling factor for the square.

    Returns
    -------
    box : np.array of shape (4, 3)
        Eukledian coordinates of the corners of the square in 3d space.
    '''
    g = geometry
    box = h2e(g['center']) + r_factor * np.vstack([h2e( g['v_y']) + h2e(g['v_z']),
                                                   h2e(-g['v_y']) + h2e(g['v_z']),
                                                   h2e(-g['v_y']) - h2e(g['v_z']),
                                                   h2e( g['v_y']) - h2e(g['v_z'])
                                               ])
    return box

def xyz_circle(geometry, r_factor=1, philim=[0, 2 * np.pi], n_vertices=90):
    '''Generate Eukledian positions along an ellipse.

    The circle is centered on the center of the object and the semi-major
    and minor axes are given by ``v_y`` and ``v_z``. Note that this function
    is usually used to generate circle position, although ellipses are possible,
    thus the name.

    The circle (or ellipse) is approximated by a polygon with ``n_vertices``
    vertices, where the value of ``n_vertices`` is taken from the ``self.display``
    dictionary.

    Parameters
    ----------
    r_factor : float
        Scaling factor for the square.
    phi_lim : list
        Lower and upper limit for the angle phi to restrict the circle to a wedge.

    Returns
    -------
    circle : np.array of shape (n, 3)
        Eukledian coordinates of the corners of the square in 3d space.
    '''
    n = n_vertices
    phi = np.linspace(0.5 * np.pi, 2.5 * np.pi, n, endpoint=False)
    v_y = r_factor * geometry['v_y']
    v_z = r_factor * geometry['v_z']

    x = np.cos(phi)
    y = np.sin(phi)
    # phi could be less then full circle
    # wrap phi at lower bound (which could be negative).
    # For the default [0, 2 pi] this is a no-op
    phi = (phi - philim[0]) % (2 * np.pi)
    ind = phi < anglediff(philim)
    x[~ind] = 0
    y[~ind] = 0

    return h2e(geometry['center'] + x.reshape((-1, 1)) * v_y + y.reshape((-1, 1)) * v_z)


class NoGeometry(object):
    _geometry = {}
    shape = 'none'

    def __init__(self, kwargs={}):
        self.geometry = self


class Geometry(NoGeometry):

    n_points = 50

    def __init__(self, kwargs={}):
        self.pos4d = _parse_position_keywords(kwargs)
        #copy class attribute to instance attribute
        self._geometry = copy(self._geometry)
        super(Geometry, self).__init__(kwargs=kwargs)

    def __getitem__(self, key):
        '''This function wraps access to the pos4d matrix.

        This is mostly a convenience method that gives access to vectors from the
        ``pos4d`` matrix in familiar terms with string labels:

        - ``center``: The ``center`` is the origin of the local coordiante system
          of the optical elemement. Typically, if will be the center of the
          active plane, e.g. the center of the mirror surface.
        - :math:`\hat v_{y,z}`: The box stretches in the y and z direction for
          :math:`\pm \hat v_y` and :math:`\pm \hat v_z`. In optics, the relevant
          interaction often happens on a surface, e.g. the surface of a mirror or
          of a reflection grating. In the defaul configuration, this surface is in
          the yz-plane.
        - :math:`\hat v_x`: The thickness of the box is not as important in many
          cases,
          but is useful to e.g. render the geometry or to test if elements collide.
          The box reaches from the yz-plane down to :math:`- \hat v_x`. Note, that
          this definition means the size parallel to the y and z axis is
          :math:`2 |\hat v_{y,z}|`, but only :math:`1 |\hat v_{x}|`.
        - :math:`\hat e_{x,y,z}`: When an object is initialized, it automatically
          adds unit vectors in the
          direction of :math:`\hat v_{x,y,z}` called :math:`\hat e_{x,y,z}`.
        - ``plane``: It also adds the homogeneous coordinates of the active plane
          as ``plane``.
        - Other labels get looked up in ``self._geometry`` and if the resulting
          value is a 4-d vector, it gets transformed with ``pos4d``.

        Not all these labels make sense for every optical element (e.g. a curved
        mirror does not really have a "plane").
        Access through this method is slower than direct indexing of ``self.pos4d``.
        '''

        if key == 'center':
            return self.pos4d[:, 3]
        elif key in ['v_x', 'e_x']:
            val = self.pos4d[:, 0]
        elif key in ['v_y', 'e_y']:
            val = self.pos4d[:, 1]
        elif key in ['v_z', 'e_z']:
            val = self.pos4d[:, 2]
        elif key == 'plane':
            return point_dir2plane(self['center'], self['e_x'])
        else:
            val = self._geometry[key]
            if isinstance(val, np.ndarray) and (val.shape[-1] == 4):
                val = np.dot(self.pos4d, val)
        if key[:2] == 'e_':
            return val / np.linalg.norm(val)
        else:
            return val

    def intersect(self, dir, pos):
        '''Calculate the intersection point between a ray and the element

        Parameters
        ----------
        dir : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the direction of the ray
        pos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of a point on the ray

        Returns
        -------
        intersect :  boolean array of length N
            ``True`` if an intersection point is found.
        interpos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the intersection point. Values are set
            to ``np.nan`` if no intersection point is found.
        interpos_local : `numpy.ndarray` of shape (N, 2)
            y and z coordinates in the coordiante system of the active plane.
        '''
        raise NotImplementedError

    def get_local_euklid_bases(self, interpos_local):
        '''Obtain a local eukledian base at a set of positions.

        Parameters
        ----------
        interpos_local : `numpy.ndarray` of shape (N, 2)
            coordinates in the coordiante system of the geometry (e.g. (x, y), or
            (r, phi)).

        Returns
        -------
        e_1, e_2, n : `numpy.ndarray` of shape (N, 4)
            Vectors pointing in direction 1, 2, and normal to the surface.
        '''
        raise NotImplementedError

class FinitePlane(Geometry):
    '''Base class for geometrically flat optical elements.

    '''

    shape = 'box'

    loc_coos_name = ['y', 'z']
    '''name for output columns that contain the interaction point in local coordinates.'''

    def intersect(self, dir, pos):
        '''Calculate the intersection point between a ray and the element

        Parameters
        ----------
        dir : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the direction of the ray
        pos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of a point on the ray

        Returns
        -------
        intersect :  boolean array of length N
            ``True`` if an intersection point is found.
        interpos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the intersection point. Values are set
            to ``np.nan`` if no intersection point is found.
        interpos_local : `numpy.ndarray` of shape (N, 2)
            y and z coordinates in the coordiante system of the active plane
            (not normalized to the dimensions of the element in question, but
            in absolute units).
        '''
        k_nominator = np.dot(self['center'] - pos, self['e_x'])
        k_denom = np.dot(dir, self['e_x'])

        is_parallel = k_denom == 0
        # To avoid warning for parallel rays, which is handeled expicitly below
        with np.errstate(divide='ignore'):
            k = k_nominator / k_denom

        forward = k >= 0  # with k < 0, rays would have to move backwards.

        if dir.ndim == 2:
            interpos = pos + k[:, None] * dir  # broadcasting array
        else:
            interpos = pos + k * dir           # dir is a scalar
        vec_center_inter = interpos - self['center']
        interpos_local = np.vstack([np.dot(vec_center_inter, self['e_y']),
                                    np.dot(vec_center_inter, self['e_z'])]).T

        intersect = (~is_parallel & forward &
                     (np.abs(interpos_local[:, 0]) <= np.linalg.norm(self['v_y'])) &
                     (np.abs(interpos_local[:, 1]) <= np.linalg.norm(self['v_z'])))
        for i in [interpos, interpos_local]:
            if dir.ndim == 2:
                i[~intersect, :3] = np.nan
            # input of single photon.
            elif (dir.ndim == 1) and not intersect:
                i[:3] = np.nan

        return intersect, interpos, interpos_local

    def get_local_euklid_bases(self, interpos_local):
        '''Obtain a local eukledian base at a set of positions.

        Parameters
        ----------
        interpos_local : `numpy.ndarray` of shape (N, 2)
            coordinates in the coordiante system of the geometry (e.g. (x, y), or
            (r, phi)).

        Returns
        -------
        e_1, e_2, n : `numpy.ndarray` of shape (N, 4)
            Vectors pointing in direction 1, 2, and normal to the surface.
        '''

        n = interpos_local.shape[0]
        x = np.tile(self['e_x'], (n, 1))
        y = np.tile(self['e_y'], (n, 1))
        z = np.tile(self['e_z'], (n, 1))
        return y, z, x


class PlaneWithHole(FinitePlane):

    shape='triangulation'
    outer_factor = 3
    inner_factor = 0

    def __init__(self, kwargs):
        self._geometry['r_inner'] = kwargs.pop('r_inner', 0.)
        super(PlaneWithHole, self).__init__(kwargs)

    def triangulate(self, display={}):
        '''Return a triangulation of the aperture hole embedded in a square.

        The size of the outer square is determined by the ``'outer_factor'`` element
        in ``self.display``.

        Returns
        -------
        xyz : np.array
            Numpy array of vertex positions in Eukeldian space
        triangles : np.array
            Array of index numbers that define triangles
        '''
        outer_disp = self.outer_display(display)
        outer_shape = self.outer_shape(display)

        xyz, triangles = plane_with_hole(outer_disp, outer_shape)

        if self['r_inner'] > 0.:
            inner_shape = self.inner_shape(display)
            # Inner edge of the display. If we have several stacked apertures,
            # we don't want to fill is all up to r=0.
            inner_disp = self.inner_display(display)
            new_xyz, new_tri = plane_with_hole(inner_shape, inner_disp)
            xyz, triangles = combine_disjoint_triangulations([xyz, new_xyz],
                                                             [triangles, new_tri])

        return xyz, triangles

    def outer_shape(self, diplay):
        raise NotImplementedError

    def outer_display(self, diplay):
        raise NotImplementedError

    def inner_shape(self, diplay):
        raise NotImplementedError

    def inner_display(self, diplay):
        raise NotImplementedError


class RectangleHole(PlaneWithHole):
    def outer_shape(self, display):
        return xyz_square(self)

    def outer_display(self, display):
        return xyz_square(self, r_factor=display['outer_factor'])


class CircularHole(PlaneWithHole):

    def __init__(self, kwargs):
        phi = kwargs.pop('phi', [0, 2. * np.pi])
        if np.max(np.abs(phi)) > 10:
            raise ValueError('Input angles >> 2 pi. Did you use degrees (radian expected)?')
        if phi[0] > phi[1]:
            raise ValueError('phi[1] must be greater than phi[0].')
        if (phi[1] - phi[0]) > (2 * np.pi + 1e-6):
            raise ValueError('phi[1] - phi[0] must be less than 2 pi.')
        self.phi = phi

        super(CircularHole, self).__init__(kwargs)

    def outer_display(self, display):
        '''Return values in Eukledian space.'''
        return xyz_circle(self, r_factor=display['outer_factor'])

    def outer_shape(self, display):
        return xyz_circle(self, r_factor=1, philim=self.phi)

    def inner_shape(self, display):
        return xyz_circle(self,
                          r_factor=self['r_inner']/np.linalg.norm(self['v_y']),
                          philim=self.phi)
    def inner_display(self, display):
        # Inner edge of the display. If we have several stacked apertures,
        # we don't want to fill is all up to r=0.
        return xyz_circle(self,
                          r_factor=display['inner_factor'] * self['r_inner'] / np.linalg.norm(self['v_y']),
                          philim=self.phi)

class Cylinder(Geometry):
    '''A Geometry shaped like a ring or tube.

    This object is shaped like a tube. The form is a circle in the xy plane
    and flat along the z-direction.  This can be used, for example to
    simulate a setup that can follow the Rowland circle
    geometry exactly which is useful, e.g. to study the resolution of a
    spectrograph without worrying about the details of the detector geometry.

    Parameters
    ----------
    position, orientation, zoom, pos4d : see description of `pos4d`
        The radius of the tube is given by the ``zoom`` keyword, see `pos4d`.
        Use ``zoom[0] == zoom[1]`` to make a circular tube.
        ``zoom[0] != zoom[1]`` gives an elliptical profile.
        ``zoom[2]`` sets the extension in the z direction.
    phi_lim : list
        If a cylinder does not cover the full circle, set ``phi_lim`` to the
        limits, e.g. ``[-np.pi / 2, np.pi / 2]`` makes a "half-pipe".
    '''
    loc_coos_name = ['phi', 'z']

    shape = 'surface'
    coos_limits = [np.array([-np.pi, np.pi]), np.array([-1, 1])]

    # This verification code does not fit here right now, but when I convert to traits
    # later it might come in useful again and thus I keep it here as a comment.
    # @property
    # def phi_lim(self):
    #     return self._phi_lim

    # @phi_lim.setter
    # def phi_lim(self, value):
    #     if (len(value) != 2) or (value[0] > value[1]):
    #         raise ValueError('phi_lim has the format [lower limit, upper limit]')
    #     for v in value:
    #         if (v < -np.pi) or (v > np.pi):
    #             raise ValueError('phi_lim must be in range -pi to +pi.')
    #     self.coos_limits[0] = value

    def __init__(self, kwargs={}):
        self.coos_limits = deepcopy(self.coos_limits)
        self.coos_limits[0] = np.asanyarray(kwargs.pop('phi_lim', [-np.pi, np.pi]))
        super(Cylinder, self).__init__(kwargs)

    def __getitem__(self, value):
        if value == 'R':
            trans, rot, zoom, shear = decompose44(self.pos4d)
            return zoom[0]
        else:
            return super(Cylinder, self).__getitem__(value)

    @classmethod
    def from_rowland(cls, rowland, width, rotation=0., kwargs={}):
        '''Generate a `Cylinder` from a `RowlandTorus`.

        According to the definition of the `marxs.design.rowland.RowlandTorus`
        the origin phi=0 is at the "top". When this class method is used to
        make a detector that catches all dispersed grating signal on the
        Rowland torus, a ``rotation=np.pi`` places the center of the Cylinder
        close to the center of the torus (the location of the focal point in
        the standard Rowland geometry).

        Parameters
        ----------
        rowland : `~marxs.design.RowlandTorus`
            The circular detector is constructed to fit exactly into the
            Rowland Circle defined by ``rowland``.
        width : float
            Half-width of the tube in the flat direction (z-axis) in mm
        rotation : float
            Rotation angle of the Cylinder around its z-axis compared to the
            phi=0 position of the Rowland torus.

        '''
        # Step 1: Rotate around z axis
        rot = axangle2mat(np.array([0, 0, 1.]), rotation)
        # Step 2: Get position and size from Rowland torus
        pos4d_circ = compose([rowland.R, 0, 0], rot, [rowland.r, rowland.r, width])
        # Step 3: Transform to global coordinate system
        pos4d_circ = np.dot(rowland.pos4d, pos4d_circ)
        # Step 4: Make detector
        det_kwargs = {'pos4d': pos4d_circ}
        det_kwargs.update(kwargs)
        return cls(det_kwargs)


    def intersect(self, dir, pos, transform=True):
        '''Calculate the intersection point between a ray and the element

        Parameters
        ----------
        dir : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the direction of the ray
        pos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of a point on the ray
        transform : bool
            If ``True``, input is in global coordinates and needs to be transformed
            here for the calculations; if ``False`` input is in local coordinates.

        Returns
        -------
        intersect :  boolean array of length N
            ``True`` if an intersection point is found.
        interpos : `numpy.ndarray` of shape (N, 4)
            homogeneous coordinates of the intersection point. Values are set
            to ``np.nan`` is no intersecton point is found.
        interpos_local : `numpy.ndarray` of shape (N, 2)
            phi, z coordiantes (in the local frame) for one of the intersection points.
            If both intersection points are required, reset ``self.inner`` and call this
            function again.
        '''
        # This could be moved to a general function
        if not np.all(dir[:, 3] == 0):
            raise ValueError('First input must be direction vectors.')
        # Could test pos, too...
        if transform:
            invpos4d = np.linalg.inv(self.pos4d)
            dir = np.dot(invpos4d, dir.T).T
            pos = np.dot(invpos4d, pos.T).T

        xyz = h2e(pos)
        dir_e = h2e(dir)

        # Solve quadratic equation in steps. a12 = (-xr +- sqrt(xr - r**2(x**2 - R**2)))
        xy = xyz[:, :2]
        r = dir[:, :2]
        c = np.sum(xy**2, axis=1) - 1.
        b = 2 * np.sum(xy * r, axis=1)
        a = np.sum(r**2, axis=1)
        underroot = b**2 - 4 * a * c
        # List of intersect in xy plane.
        intersect = (underroot >= 0)
        i = intersect  # just a shorthand because it's used so much below

        interpos_local = np.ones((pos.shape[0], 2))
        interpos_local[:] = np.nan
        interpos = np.ones_like(pos)
        interpos[:] = np.nan

        if intersect.sum() > 0:
            i_ind = intersect.nonzero()[0]
            denom = 2 * a[i]
            a1 = (- b[i] + np.sqrt(underroot[i])) / denom
            a2 = (- b[i] - np.sqrt(underroot[i])) / denom
            xy_1 = xy[i, :] + a1[:, np.newaxis] * r[i, :]
            phi_1 = np.arctan2(xy_1[:, 1], xy_1[:, 0])
            xy_2 = xy[i, :] + a2[:, np.newaxis] * r[i, :]
            phi_2 = np.arctan2(xy_2[:, 1], xy_2[:, 0])
            # 1, 2 look like hits in x,y but might still miss in z
            z_1 = xyz[i, 2] + a1 * dir[i, 2]
            z_2 = xyz[i, 2] + a2 * dir[i, 2]
            hit_1 = ((a1 >= 0) & (np.abs(z_1) <= 1.) &
                     angle_between(phi_1, self.coos_limits[0][0], self.coos_limits[0][1]))
            hit_2 = ((a2 >= 0) & (np.abs(z_2) <= 1.) &
                     angle_between(phi_2, self.coos_limits[0][0], self.coos_limits[0][1]))
            # If both 1 and 2 are hits, use the closer one
            hit_1[hit_2 & (a2 < a1)] = False
            hit_2[hit_1 & (a2 >= a1)] = False
            intersect[i_ind] = hit_1 | hit_2
            # Set values into array from either point 1 or 2
            interpos_local[i_ind[hit_1], 0] = phi_1[hit_1]
            interpos_local[i_ind[hit_1], 1] = z_1[hit_1]

            interpos_local[i_ind[hit_2], 0] = phi_2[hit_2]
            interpos_local[i_ind[hit_2], 1] = z_2[hit_2]
            # Calculate pos for point 1 or 2 in local xyz coord system
            interpos[i_ind[hit_1], :] = e2h(xyz[i_ind, :] + a1[:, None] * dir_e[i_ind, :], 1)[hit_1, :]
            interpos[i_ind[hit_2], :] = e2h(xyz[i_ind, :] + a2[:, None] * dir_e[i_ind, :], 1)[hit_2, :]


            trans, rot, zoom, shear = decompose44(self.pos4d)
            # interpos_local in z direction is in local coordinates, i.e.
            # the x coordiante is 0..1, but we want that in units of the
            # global coordinate system.
            interpos_local[:, 1] = interpos_local[:, 1] * zoom[2]
            interpos = np.dot(self.pos4d, interpos.T).T

        return intersect, interpos, interpos_local


    def parametric_surface(self, phi=None, z=None, display={}):
        '''Parametric description of the tube.

        This is just another way to obtain the shape of the tube, e.g.
        for visualization.

        Parameters
        ----------
        phi : np.array
            ``phi`` is the angle around the tube profile. Set to ``None`` to use the
            extend of the element itself.
        z : np.array
            The coordiantes along the radius coordinate. Set to ``None`` to use the
            extend of the element itself.

        Returns
        -------
        xyzw : np.array
            Ring coordinates in global homogeneous coordinate system.
        '''
        phi = np.linspace(self.coos_limits[0][0], self.coos_limits[0][1], self.n_points) \
              if phi is None else np.asanyarray(phi)
        trans, rot, zoom, shear = decompose44(self.pos4d)
        z = self.coos_limits[1] if z is None else np.asanyarray(z) / zoom[2]
        if (phi.ndim != 1) or (z.ndim != 1):
            raise ValueError('input parameters have 1-dim shape.')
        phi, z = np.meshgrid(phi, z)
        x = np.cos(phi)
        y = np.sin(phi)
        w = np.ones_like(z)
        coos = np.array([x, y, z, w]).T
        return np.einsum('...ij,...j', self.pos4d, coos)

    def get_local_euklid_bases(self, interpos_local):
        '''Obtain a local eukledian base at a set of positions.

        Parameters
        ----------
        interpos_local : `numpy.ndarray` of shape (N, 2)
            coordinates in the coordiante system of the geometry (e.g. (x, y), or
            (r, phi)).

        Returns
        -------
        e_1, e_2, n : `numpy.ndarray` of shape (N, 4)
            Vectors pointing in direction 1, 2, and normal to the surface.
        '''

        phi = interpos_local[:, 0]
        zeros = np.zeros_like(phi)
        e_phi = np.array([-np.sin(phi), np.cos(phi), zeros, zeros]).T
        e_z = np.tile(self['e_z'], (interpos_local.shape[0], 1))
        e_n = np.array([np.cos(phi), np.sin(phi), zeros, zeros]).T

        e_phi = np.einsum('...ij,...j', self.pos4d, e_phi)
        e_n = np.einsum('...ij,...j', self.pos4d, e_n)
        # e_z is already in the local coordinate system
        return norm_vector(e_phi), e_z, norm_vector(e_n)



# Licensed under GPL version 3 - see LICENSE.rst
'''Gratings and efficiency files'''
import math
import numpy as np
from numpy.core.umath_tests import inner1d

from ..math.utils import norm_vector, h2e, e2h
from ..math.polarization import parallel_transport
from .. import energy2wave
from .base import FlatOpticalElement


class OrderSelector(object):
    '''Select from a list of order numbers independent of energy.

    Parameters
    ----------
    orderlist : array
        This are the order numbers to chose from. They must be integers.
    p : None or array
        Probability for a photon to end up in each order. The sum of all probabilities
        can be smaller than 1, if a certain fraction of photons is absorbed by
        the grating. If this is ``None``, equal probability is assigned to all orders.

    Examples
    --------
    Two common use cases for testing are a grating where every photon gets
    diffracted into the same order or where all photons get distributed with
    equal probability into a set of orders.

    >>> import numpy as np
    >>> from marxs.optics import FlatGrating, OrderSelector
    >>> singleordergrating = FlatGrating(d=1e-4, order_selector=OrderSelector([3]))
    >>> setofordersgrating = FlatGrating(d=1e-4, order_selector=OrderSelector(np.arange(-3, 4)))
    '''
    def __init__(self, orderlist, p=None):
        self.orderlist = orderlist
        if p is None:
            self.p = np.ones(len(orderlist)) / len(orderlist)
        else:
            p = np.asanyarray(p)
            if len(p) != len(orderlist):
                raise ValueError('Number of elements in orderlist and probabilities does not match.')
            if p.sum() > 1.:
                raise ValueError('Sum of all probabilities must be <= 1.')
            if np.any(p < 0):
                raise ValueError('Probabilities cannot be negative')
            self.p = p

    def __call__(self, energy, *args):
        p_sum = self.p.sum()
        p = self.p / p_sum
        if np.isscalar(energy):
            return np.random.choice(self.orderlist, p=p), self.p.sum()
        else:
            return np.random.choice(self.orderlist, size=len(energy), p=p), p_sum * np.ones_like(energy)


class EfficiencyFile(object):
    '''Select grating order from a probability distribution in a data file.

    The file format supported by this class is as follows:
    The first colum contains energy values in keV, all remaining columns have the probability
    that a photons with this energy is diffracted into the respective order. The probabilities
    for each order do not have to add up to 1.

    Parameters
    ----------
    filename : string
        Path to the efficiency file.
    orders : list
        List of orders in the file. Must match the number of columns with probabilities.
    '''
    def __init__(self, filename, orders):
        dat = np.loadtxt(filename)
        self.energy = dat[:, 0]
        if len(orders) != (dat.shape[1] - 1):
            raise ValueError('orders has len={0}, but data files has {1} order columns.'.format(len(orders), dat.shape[1] - 1))
        self.orders = np.array(orders)
        self.prob = dat[:, 1:]
        # Probability to end up in any order
        self.totalprob = np.sum(self.prob, axis=1)
        # Cumulative probability for orders, normalized to 1.
        self.cumprob = np.cumsum(self.prob, axis=1) / self.totalprob[:, None]

    def __call__(self, energies, *args):
        orderind = np.empty(len(energies), dtype=int)
        ind = np.empty(len(energies), dtype=int)
        for i, e in enumerate(energies):
            ind[i] = np.argmin(np.abs(self.energy - e))
            orderind[i] = np.min(np.nonzero(self.cumprob[ind[i]] > np.random.rand()))
        return self.orders[orderind], self.totalprob[ind]



class FlatGrating(FlatOpticalElement):
    '''Flat grating

    The grating is assumed to be geometrically thin, i.e. all photons enter on
    the face of the grating, not through the sides.

    Parameters
    ----------
    d : float or callable
        grating constant in mm. If ``d`` is callable, then it will be called as
        ``d(intercoos)``, where ``intercoos`` is a (N, 2) array holding the
        positions where photons hit the gratings in the local coordinate system.
        This can be used to simulate manufacturing uncertainties or intentional
        grating period variation. The callable has to return a vector of lengths N
        that contains the grating constant for each photon in mm.
    order_selector : callable
        A function or callable object that accepts arrays of photon
        energy, polarization and the blaze angle as input and returns
        arrays for grating order (integer) and probability
        (float). The probabiliy expresses the chance that the photon
        passes the grating and is not absorbed, e.g. if the
        probability that a photon at energy E ends up in order=[-2,
        -1, 0, 1, 2] is [0, 0, .5, .3, .0] , then the returned
        probability for all photons should be 0.8.
    transmission : bool
        Set to ``True`` for a transmission grating and to ``False`` for a
        reflection grating. (*Default*: ``True`` ) **Warning: Reflection
        gratings have not yet been tested**
    groove_angle : float
        Angle between the local z axis and the direction of the
        grooves in radian.  (*Default*: ``0.``)
    '''

    loc_coos_name = ['grat_y', 'grat_z']
    '''name for output columns that contain the interaction point in local coordinates.'''

    order_name = 'order'
    '''Column name for diffraction order. If set to ``None`` this is not added to the photon list.'''

    blaze_name = 'blaze'
    '''Column name for blaze angle. If set to ``None`` this is not added to the photon list.'''

    def order_sign_convention(self, p, e_perp_groove):
        '''Set sign convention for grating orders.

        This sets the following, somewhat arbitrary convention:
        Positive grating orders will are displaced along the local
        :math:`\hat e_z` vector, negative orders in the opposite
        direction. If the grating is rotated by :math:`-\pi` the
        physical situation is the same, but the sign of the grating
        order will be reversed.  In this sense, the convention chosen
        is arbitrarily. However, it has some practical advantages: The
        implementation is fast and all photons passing through the
        grating in the same diffraction order are displaced in the
        same way. (Contrary to the convention in :class:`CATGrating`.)
        ``order_sign_convention`` has to be a callable that accepts an
        array of eukledian direction vectors as input and returns
        ``+1``, ``-1``, or an array filled with ``-1`` or ``+1``.

        Parameters
        ----------
        p : np.array
            Array of Eucleadian direction vectors
        e_perp_groove : np.array
            Array of local groove directions at the positions where the photons hit
        '''
        # -1 because n, l, d should be right-handed coordinate system
        # while n = e_x, l = e_x, and d = e_y would be left-handed.
        return -1

    def __init__(self, **kwargs):
        self.order_selector = kwargs.pop('order_selector')
        self.transmission = kwargs.pop('transmission', True)
        if 'd' not in kwargs:
            raise ValueError('Input parameter "d" (Grating constant) is required.')
        self._d = kwargs.pop('d')
        groove_angle = kwargs.pop('groove_angle', 0.)

        super(FlatGrating, self).__init__(**kwargs)

        self.geometry._geometry['groove_angle'] = groove_angle

    def e_groove_coos(self, intercoos):
        '''Get coordiante orthonormal coordinate system along groove direction.

        Parameters
        ----------
         intercoos : `numpy.ndarray` of shape (N, 2)
            coordinates in the coordiante system of the geometry (e.g. (x, y), or
            (r, phi)).

        Returns
        -------
        e_groove, e_perp_groove, n : `numpy.ndarray` of shape (N, 4)
            Vectors pointing along the groove direction (parallel to the surface),
            perpendicular to the groove direction (parallel to surface),
            and normal to surface.
        '''

        groove = self.geometry['groove_angle']
        ex, ey, en = self.geometry.get_local_euklid_bases(intercoos)
        e_groove = math.sin(-groove) * ex + math.cos(-groove) * ey
        e_perp_groove = math.cos(-groove) * ex - math.sin(-groove) * ey
        return e_groove, e_perp_groove, en

    def d(self, intercoos):
        '''Method that returns the grating constant at given positions.

        For a grating with constant grating constant, this will just return the
        number input as ``d`` when the element was initialized. For gratings
        where the grating constant varies with position on the facet, this
        calculates the appropriate number for every position.
        '''
        if not callable(self._d):
            return self._d
        else:
            return self._d(intercoos)

    def blaze_angle_modifier(self, intercoos):
        '''Modify blaze angle

        In `diffract_photons` the blaze angle is calculated relative to the surface
        of the grating. In cases where that number has to be modified (e.g. when the grating
        bars are not perpendicualr to the surface) the blaze angle can be modified by
        overriding this function.
        '''
        return np.zeros(intercoos.shape[0])

    def diffract_photons(self, photons, intersect, interpos, intercoos):
        '''Vectorized implementation'''
        p = norm_vector(photons['dir'].data[intersect])
        l, d, n = self.e_groove_coos(intercoos[intersect])
        # Minus sign here because we want n, l, d to be a right-handed coordinate system
        d = - d

        wave = energy2wave / photons['energy'].data[intersect]
        # calculate angle between normal and (ray projected in plane perpendicular to groove)
        # -> this is the blaze angle
        p_perp_to_grooves = norm_vector(p - inner1d(p, l)[:, None] * l)
        # Use abs here so that blaze angle is always in 0..pi/2
        # independent of the relative orientation of p and n.
        blazeangle = np.arccos(np.abs(inner1d(p_perp_to_grooves, n)))
        blazeangle += self.blaze_angle_modifier(intercoos[intersect, :])
        m, prob = self.order_selector(photons['energy'].data[intersect],
                                      photons['polarization'].data[intersect],
                                      blazeangle)

        # The idea to calculate the components in the (d,l,n) system separately
        # is taken from MARX
        sign = self.order_sign_convention(p, d)
        p_d = inner1d(p, d) + sign * m * wave / self.d(intercoos[intersect, :])
        p_l = inner1d(p, l)
        # The norm for p_n can be derived, but the direction needs to be chosen.
        p_n = np.sqrt(1. - p_d**2 - p_l**2)
        # Check if the photons have same direction compared to normal before
        direction = np.sign(inner1d(p, n), dtype=np.float)
        if not self.transmission:
            direction *= -1
        dir = p_d[:, None] * d + p_l[:, None] * l + (direction * p_n)[:, None] * n
        return dir, m, prob, blazeangle

    def specific_process_photons(self, photons, intersect, interpos, intercoos):

        dir, m, p, blaze = self.diffract_photons(photons, intersect, interpos, intercoos)
        pol = parallel_transport(photons['dir'].data[intersect, :], dir,
                                 photons['polarization'].data[intersect, :])
        out = {'dir': dir, 'probability': p, 'polarization': pol}
        if self.order_name is not None:
            out[self.order_name] = m
        if self.blaze_name is not None:
            out[self.blaze_name] = blaze
        return out


class CATGrating(FlatGrating):
    '''Critical-Angle-Transmission Grating

    CAT gratings are a special case of :class:`FlatGrating` and accept the same arguments.

    They differ from a :class:`FlatGrating` in the sign convention of the
    grating orders: Blazing happens on the side of the negative orders. Obviously, this
    convention is only meaningful if the photons do not arrive perpendicular to the grating.
    '''


    def order_sign_convention(self, p, e_perp_groove):
        '''Convention to chose the sign for CAT grating orders

        Blazing happens on the side of the negative orders. Obviously, this
        convention is only meaningful if the photons do not arrive perpendicular to the grating.
        '''
        dotproduct = inner1d(p, e_perp_groove)
        sign = np.sign(dotproduct)
        sign[sign == 0] = 1
        return sign



# Licensed under GPL version 3 - see LICENSE.rst
import pytest
import numpy as np
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename
from transforms3d.euler import euler2mat

from ..catgrating import *
from ..catgrating import DataFileFormatException, check_lx_dims
from marxs.optics import CATGrating, OrderSelector, FlatDetector
from marxs.optics.scatter import RandomGaussianScatter
from marxs.utils import generate_test_photons

def test_nonparallelCATGrating_simplifies_to_CATGrating():
    '''With all randomness parameters set to 0, the results
    should match a CAT grating exactly.'''
    photons = generate_test_photons(50)
    # Mix the photons up a little
    scat = RandomGaussianScatter(scatter=0.01 * u.rad)
    photons = scat(photons)
    order_selector = OrderSelector([2])
    cat = CATGrating(order_selector=order_selector, d=0.001)
    npcat = NonParallelCATGrating(order_selector=order_selector, d=0.001)
    p1 = cat(photons.copy())
    p2 = npcat(photons)
    assert np.all(p1['dir'] == p2['dir'])

def test_check():
    '''Make sure check raises Exception for unphysical parameters'''
    with pytest.raises(ValueError):
        check_lx_dims({'barwidth': 1, 'period': 1})

def test_scalingSitransparancy():
    '''The module has data for 1 mu Si and scales that to the depth of the
    grating. Test against known-good coefficient from CXRO.'''
    photons = generate_test_photons(100)
    cat = L1(order_selector=OrderSelector([-5]),
             l1_dims = {'bardepth': 0.004 * u.mm,
                        'period': 0.0001 * u.mm,
                        'barwidth': 0.00009 * u.mm})
    photons = cat(photons)
    assert np.isclose(np.median(photons['probability']), 0.22458, rtol=1e-4)


def test_L2_broadening():
    '''Compare calculated L2 broadening with the values calculated
    by Ralf Heilmann. Both calculation make simplications (e.g. assuming
    a circular hole), but should give the right order of magnitude for
    diffraction.'''
    photons = generate_test_photons(1000)
    photons['energy'] = 0.25  # 50 Ang
    l2 = L2Diffraction(l2_dims={'period': 0.966 * u.mm,
                                'barwidth': 0.1 * u.mm})
    photons = l2(photons)
    det = FlatDetector(position=[-1, 0, 0])
    photons = det(photons)
    sigma = np.std(photons['det_x'])
    # for small angles tan(alpha) = alpha
    # 1 arcsec = 4.8...e-6 rad
    # 0.5 is sigma calculated by Ralf Heilmann for a rectangular
    # aperture
    assert np.isclose(sigma, .5 * 4.8e-6, rtol=.5)


def test_L2_Abs():
    '''Check L2 absorption against numbers calculated by Ralf Heilmann.'''
    photons = generate_test_photons(1)
    l2 = L2Abs(l2_dims={'period': 0.966 * u.mm, 'bardepth': 0.5 * u.mm,
                        'barwidth': 0.1 * u.mm})
    photons = l2(photons)
    assert np.isclose(photons['probability'], 0.81, rtol=0.02)


def test_L2_Abs_angle():
    '''Test that that shadow area increases with angle.
    The comparison number was calculated by H. M. Guenther in response
    to Arcus DQ36.'''
    l2_dims = {'period': 0.966 * u.mm, 'bardepth': 0.5 * u.mm,
               'barwidth': 0.1 * u.mm}

    l2 = L2Abs(l2_dims=l2_dims)
    p1 = l2(generate_test_photons(1))

    l2 = L2Abs(l2_dims=l2_dims, orientation=euler2mat(np.deg2rad(1.8), 0, 0, 'szxy'))
    p2 = l2(generate_test_photons(1))

    assert np.isclose(p2['probability'][0] / p1['probability'][0], 0.979, rtol=1e-3)


def test_efficiency_table():
    '''Test that the efficiency is read in in the right format.'''
    efftab = InterpolateEfficiencyTable(get_pkg_data_filename('grating_efficiency.csv'), k=2)
    orders, interpprobs = efftab.probabilities(np.array([0.5, 0.5, 1, 1]),
                                               np.ones(4),
                                               np.deg2rad([1., 2., 1., 2.]))
    assert np.allclose(interpprobs.sum(axis=0), [.8, .8, 1., 1.], rtol=1e-4)
    assert np.allclose(interpprobs[:, 0], [.3, .3, .2], rtol=1e-4)


def test_efficiency_table_in_use():
    '''Use table in a optical element'''
    efftab = InterpolateEfficiencyTable(get_pkg_data_filename('grating_efficiency.csv'), k=2)
    cat = CATGrating(order_selector=efftab, d=0.001)
    photons = generate_test_photons(5000)
    photons = cat(photons)
    assert np.isclose((photons['order']==0).sum(), len(photons) / 2, rtol=.05)


def test_efficiency_table_wrong_format():
    '''Try to load a datafile with mission rows.'''
    with pytest.raises(DataFileFormatException):
        efftab = InterpolateEfficiencyTable(get_pkg_data_filename('grating_efficiency_broken.csv'), k=2)



# Licensed under GPL version 3 - see LICENSE.rst
from io import StringIO
import numpy as np
from numpy.random import random
from astropy.table import Table
from transforms3d import axangles, affines
import pytest

from ..grating import (FlatGrating, CATGrating,
                       OrderSelector, EfficiencyFile)
from ...math.utils import h2e
from ...math.polarization import polarization_vectors
from ... import energy2wave
from ...utils import generate_test_photons

def test_zeros_order():
    '''Photons diffracted into order 0 should just pass through'''
    dir = random((10, 4))
    dir[0, :] = dir[0, :] * 10  # make sure mostion is mostly along x so they hit grating
    photons = Table({'pos': random((10, 4)) - 1,
                     'dir': dir,
                     'energy': random(10),
                     'polarization': polarization_vectors(dir, np.random.rand(10)),
                     'probability': np.ones(10),
                     })
    # Make sure homogeneous coordiantes are valid
    # and infall is normal
    photons['pos'][:, 3] = 1
    photons['dir'][:, 1:] = 0

    p = photons.copy()
    g0 = FlatGrating(d=1./500.,
                     order_selector=OrderSelector([0]),
                     zoom=np.array([1., 50., 50.]))
    p = g0(p)
    # Direction unchanged
    d_in = h2e(photons['dir'])
    d_out = h2e(p['dir'])
    # normalize
    d_in =  d_in / np.sqrt(np.sum(d_in**2, axis=-1))[:, None]
    d_out =  d_out / np.sqrt(np.sum(d_out**2, axis=-1))[:, None]
    assert np.allclose(d_in, d_out)
    # all intersection points in y-z plane
    assert np.allclose(p['pos'][:, 0], 0.)
    # no offset between old and new ray
    assert np.allclose(np.cross(h2e(photons['pos']) - h2e(p['pos']), h2e(p['dir'])), 0)

def test_translation_invariance():
    '''For homogeneous gratings, the diffraction eqn does not care where the ray hits.'''
    photons = generate_test_photons(2)
    photons['dir'] = np.tile(np.array([-1., 2., 3., 0.]), (2, 1))
    photons['polarization'] = polarization_vectors(photons['dir'], [2., 4.])
    pos = np.tile(np.array([1., 0., 0., 1.]), (2, 1))
    delta_pos = np.array([0, .23, -.34, 0.])
    pos[1,:] += delta_pos
    photons['pos'] = pos
    for order in [-1, 0, 1]:
        g = FlatGrating(d=1./500, order_selector=OrderSelector([order]), zoom=20)
        p = g(photons.copy())
        assert np.all(p['order'] == order)
        assert np.allclose(p['pos'][:, 0], 0)
        assert np.allclose(p['dir'][0, :], p['dir'][1, :])
        assert np.allclose(p['pos'][0, :], p['pos'][1, :] - delta_pos)

def test_angle_dependence():
    '''Test the grating angle for non-perpendicular incidence.

    For this test, I write the grating equation in its simple traditional form:
    :math:`m \lambda = \delta s = d sin(\alpha)`.

    If the incoming ray is not perpendicular to the grating, we use the angle beta
    between the incoming ray and the grating normal. The grating equation than looks
    like this for m = 1:
    :math:`\delta s = \delta s_2 - \delta s_1 = d sin(\alpha) - d sin(\beta)`.
    '''
    d = 0.001
    beta = np.deg2rad([0., 5., 10., 20.])
    dir = np.zeros((len(beta), 4))
    dir[:, 1] = np.sin(beta)
    dir[:, 0] = -np.cos(beta)
    pos = np.ones((len(beta), 4))
    photons = Table({'pos': pos,
                     'dir': dir,
                     'energy': np.ones(len(beta)),
                     'polarization': polarization_vectors(dir, [1, 2, 3, 4]),
                     'probability': np.ones(len(beta)),
                     })
    g = FlatGrating(d=0.001,  order_selector=OrderSelector([1]), zoom=20)
    p = g(photons)
    # compare sin alpha to expected value: sin(alpha) = sin(beta) + lambda/d
    alpha = np.arctan2(p['dir'][:, 1], -p['dir'][:, 0])
    assert np.allclose(np.sin(alpha), np.sin(beta) + 1.2398419292004202e-06/d)

def test_order_dependence():
    '''For small theta, the change in direction is m * dtheta'''
    photons = Table({'pos': np.ones((5,4)),
                     'dir': np.tile([-1., 0, 0, 0], (5,1)),
                     'energy': np.ones(5),
                     'polarization': np.tile([0.,1.,0.,0.], (5,1)),
                     'probability': np.ones(5),
                     })
    def mock_order(x, y, z):
        return np.array([-2, -1, 0, 1, 2]), np.ones(5)
    g = FlatGrating(d=1./500, order_selector=mock_order)
    p = g(photons)
    # grooves run in z direction
    assert np.allclose(p['dir'][:, 2], 0.)
    # positive and negative orders are mirrored
    assert np.allclose(p['dir'][[0,1], 1], - p['dir'][[4,3], 1])
    # order 2 is twice as far as order 1 (for theta  small)
    assert np.abs(p['dir'][4, 1] / p['dir'][3, 1] - 2 ) < 0.00001

def test_energy_dependence():
    '''The grating angle should depend on the photon wavelength <-> energy.'''
    photons = Table({'pos': np.ones((5,4)),
                     'dir': np.tile([-1., 0, 0, 0], (5,1)),
                     'energy': np.arange(1., 6),
                     'polarization': np.tile([0, 1., 0., 0.], (5, 1)),
                     'probability': np.ones(5),
                     })
    g = FlatGrating(d=1./500, order_selector=OrderSelector([1]))
    p = g(photons)
    # grooves run in z direction
    assert np.allclose(p['dir'][:, 2], 0.)
    # n lambda = d sin(theta)
    lam = energy2wave / p['energy']
    theta = np.arctan2(p['dir'][:, 1], p['dir'][:, 0])
    assert np.allclose(1. * lam, 1./500. * np.sin(theta))

def test_blaze_dependence():
    '''Ensure that the blaze angle is passed to the order selector (regression test)'''
    def order_selector(energy, polarization, blaze):
        '''test function (unphysical)'''
        return np.floor(blaze * 10), np.ones_like(energy)

    photons = generate_test_photons(2)
    photons['dir'][1, :] = [-1,1, 0, 0]
    photons['polarization'][1, :] = [0, 0, 1, 0]
    g = FlatGrating(d=1., order_selector=order_selector, zoom=5)
    p = g(photons)
    assert np.allclose(p['order'], [0, 7])

def test_groove_direction():
    '''Direction of grooves may not be parallel to z axis.'''
    photons = generate_test_photons(5)
    order1 = OrderSelector([1])

    g = FlatGrating(d=1./500, order_selector=order1)
    e_groove, e_perp_groove, n = g.e_groove_coos(np.ones((2,1)))
    assert np.isclose(np.dot(e_groove[0, :], e_perp_groove[0, :]), 0.)
    p = g(photons.copy())

    g1 = FlatGrating(d=1./500, order_selector=order1, groove_angle=.3)
    p1 = g1(photons.copy())

    pos3d = axangles.axangle2mat([1,0,0], .3)
    g2 = FlatGrating(d=1./500, order_selector=order1, orientation=pos3d)
    p2 = g2(photons.copy())

    pos4d = axangles.axangle2aff([1,0,0], .1)
    g3 = FlatGrating(d=1./500, order_selector=order1, groove_angle=.2, pos4d=pos4d)
    p3 = g3(photons.copy())


    def angle_in_yz(vec1, vec2):
        '''project in the y,z plane (the plane of the grating) and calculate angle.'''
        v1 = vec1[1:3]
        v2 = vec2[1:3]
        arccosalpha = np.dot(v1, v2) / np.sqrt(np.dot(v1, v1) * np.dot(v2, v2))
        return np.arccos(arccosalpha)

    assert np.allclose(angle_in_yz(p1['dir'][0,:], p2['dir'][0, :]), 0, atol=5e-8)
    assert np.allclose(angle_in_yz(p3['dir'][0,:], p2['dir'][0, :]), 0, atol=5e-8)

    for px in [p1, p2, p3]:
        assert np.allclose(angle_in_yz(p['dir'][0,:], px['dir'][0, :]), 0.3)


def test_order_convention():
    dirs = np.zeros((3, 4))
    dirs[1, 2] = 0.01
    dirs[2, 2] = -0.01
    dirs[:, 0] = - 1
    photons = Table({'pos': np.ones((3, 4)),
                     'dir': dirs,
                     'energy': np.ones(3),
                     'polarization': polarization_vectors(dirs, np.ones(3)),
                     'probability': np.ones(3),
                     })
    gp = FlatGrating(d=1./500, order_selector=OrderSelector([1]), zoom=2)
    p1 = gp(photons.copy())
    gm = FlatGrating(d=1./500, order_selector=OrderSelector([-1]), zoom=2)
    m1 = gm(photons.copy())
    assert np.all(p1['order'] == 1)
    assert np.all(m1['order'] == -1)
    # intersection point with grating cannot depend on order
    assert np.all(p1['pos'].data == m1['pos'].data)


def test_CAT_order_convention():
    dirs = np.array([[-1, 0., 0., 0],
                     [-1, 0.01, -0.01, 0],
                     [-1, 0.01, 0.01, 0],
                     [-1, -0.01, 0.01, 0],
                     [-1, -0.01, -0.01, 0]])
    photons = Table({'pos': np.ones((5, 4)),
                     'dir': dirs,
                     'energy': np.ones(5),
                     'polarization': polarization_vectors(dirs, np.arange(5)),
                     'probability': np.ones(5),
                     })
    gp = CATGrating(d=1./5000, order_selector=OrderSelector([5]), zoom=2)
    p5 = gp(photons.copy())
    gm = CATGrating(d=1./5000, order_selector=OrderSelector([-5]), zoom=2)
    m5 = gm(photons.copy())
    for g in [gm, gp]:
        e_groove, e_perp, n = g.e_groove_coos(np.zeros((5, 2)))
        signs = g.order_sign_convention(photons['dir'], e_perp)
        assert np.all(signs == np.array([1, 1, 1, -1, -1]))
    assert np.all(p5['dir'][1:3, 1] > 0)
    assert np.all(p5['dir'][3:, 1] < 0)
    assert np.all(m5['dir'][1:3, 1] < 0)
    assert np.all(m5['dir'][3:, 1] > 0)


def test_uniform_efficiency():
    '''Uniform efficiency should give results from -order to +order'''
    f = OrderSelector(np.arange(-2, 3))
    # Make many numbers, to reduce chance of random failure for this test
    testout = f(np.ones(10000), np.ones(10000))
    assert set(testout[0]) == set([-2, -1 , 0, 1, 2])
    assert len(testout[0]) == len(testout[1])


def test_OrderSelector_probabilities():
    '''Most tests here use a constant order. This test makes sure that the
    probability is properly used if passed in.
    '''
    f = OrderSelector([0,1], [.3, .6])
    # Make many numbers, to reduce chance of random failure for this test
    testout = f(np.ones(10000), np.ones(10000))
    assert set(testout[0]) == set([0, 1])
    assert np.allclose(testout[1], 0.9)
    assert (testout[0] == 0).sum() < 0.8 * (testout[0] == 1).sum()


def test_OrderSelector_failures():
    with pytest.raises(ValueError) as e:
        f = OrderSelector([1, 2, 3, 4], p=[0.1, 0.2, 0.3])
    assert 'does not match' in str(e.value)

    with pytest.raises(ValueError) as e:
        f = OrderSelector([1, 2, 3], p=[1, 0.2, 0.3])
    assert 'must be <= 1' in str(e.value)

    with pytest.raises(ValueError) as e:
        f = OrderSelector([1, 2, 3], p=[-0.1, 0.2, 0.3])
    assert 'negative' in str(e.value)


def test_constant_order():
    '''Constant order will give always the same order'''
    f = OrderSelector([2])
    testout = f(np.ones(15), np.ones(15))
    assert np.all(testout[0] == 2)
    assert len(testout[0]) == len(testout[1])


def test_EfficiencyFile():
    '''Read in the efficency from a file.

    To simplify testing, simulate an inout file using StringIO.
    '''
    data  = StringIO(u".5 .1 .1 .1 .4\n1. .1 .1 .1 .5\n1.5 0. .1 .0 .5")
    eff = EfficiencyFile(data, [1, 0, -1, -2])
    testout = eff(np.array([.7, 1.1]), np.zeros(2))
    assert np.allclose(testout[1], np.array([.7, .8]))

    testout = eff(1.6 * np.ones(1000), np.zeros(1000))
    assert np.allclose(testout[1], .6)
    assert set(testout[0]) == set([-2 , 0])
    assert (testout[0] == 0).sum()  < (testout[0] == -2).sum()


def test_CATGRating_misses():
    '''Regression test: CAT gratings that intersect only a fraction of rays
    returned an array of the wrong dimension from order_sign_convention.'''
    photons = generate_test_photons(5)
    photons['pos'][:, 1] = np.arange(5)

    cat = CATGrating(d=1./5000, order_selector=OrderSelector([5]), zoom=2)
    p = cat(photons)
    assert np.all(np.isnan(p['order'][3:]))
    assert np.all(np.isnan(p['grat_y'][3:]))


def test_grating_d_callable():
    '''d can be a function.'''
    photons = generate_test_photons(5)
    photons['pos'][0, 1] = 1.
    photons['pos'][1, 1] = -1.

    def dfunc(intercoos):
        darr = np.ones(intercoos.shape[0]) * 2e-4
        d = np.where(intercoos[:, 0]>=0, darr, darr / 2.)
        return d

    grat = FlatGrating(order_selector=OrderSelector([1]), zoom=2, d=dfunc)
    p = grat(photons)
    # negative one has half the grating constant and thus twice the angle
    assert np.abs(p['dir'][1, 1] / p['dir'][0, 1] - 2 ) < 0.00001

def test_change_position_after_init():
    '''Regression: In MARXS 0.1 the geometry could not be changed after init
    because a lot of stuff was pre-calculated in __init__. That should no
    longer be the case. This test is here to check that for gratings.'''
    photons = generate_test_photons(5)

    g1 = FlatGrating(d=1./500, order_selector=OrderSelector([1]), groove_angle=.3)
    g1.order_name = 'one'
    p1 = g1(photons.copy())

    pos3d = axangles.axangle2mat([1,0,0], .3)
    trans = np.array([0., -.3, .5])
    # Make grating at some point
    g2 = FlatGrating(d=1./500, order_selector=OrderSelector([1]), position=trans)
    # then move it so that pos and groove direction match g1
    g2.geometry.pos4d = affines.compose(np.zeros(3), pos3d, 25.*np.ones(3))
    g2.blaze_name = 'myblaze'
    p2 = g2(photons.copy())

    assert np.allclose(p1['dir'], p2['dir'])
    assert np.allclose(p1['pos'], p2['pos'])

    # Check naming of the grating output columns
    assert np.all(p1['one'] == 1)
    assert 'myblaze' in p2.colnames


def test_gratings_are_independent():
    '''Regression test: Some grating properties are stored in a dict.
    This test ensures that the relevant numbers are in an instance attribute and not
    in a class attribute.
    However, we want to test only user visible properties, not hidden dicts like
    FlatGrating._geometry. So, compare grove dirs for two gratings.
    '''
    g1 = FlatGrating(d=1./500, order_selector=OrderSelector([1]), groove_angle=.3)
    g2 = FlatGrating(d=1./500, order_selector=OrderSelector([1]))
    groove1, perp1, n1 = g1.e_groove_coos(np.array([[-1, 2.3]]))
    groove2, perp2, n2 = g2.e_groove_coos(np.array([[-1, 2.3]]))
    assert not np.allclose(groove1, groove2)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from ..polarization import polarization_vectors, paralleltransport_matrix
from .. import utils

def test_random_polarization():
    '''tests the angles to vector function for polarization

    This test makes sure that there is no obvious bias in a single direction.
    *** NOTE *** It is possible, but extremely unlikely, for this test to fail by chance.
    TODO: Perform a statistical calculation to show this test is reasonable.
    TODO: Test for other possible issues (ex: polarization in +/-x axis would not be detected)
    '''
    v_1 = np.random.uniform(size=3)
    v_1 /= np.linalg.norm(v_1)

    n = 100000

    dir_array = np.tile(v_1, (n, 1))
    angles = np.random.uniform(0, 2 * np.pi, n)

    polarization = polarization_vectors(dir_array, angles)

    assert np.allclose(polarization, polarization / np.linalg.norm(polarization, axis=1)[:, np.newaxis])

    x = sum(polarization[:, 0])
    y = sum(polarization[:, 1])
    z = sum(polarization[:, 2])

    v_2 = np.array([x, y, z])

    assert np.isclose(np.dot(v_1, v_2), 0)

    assert np.linalg.norm(v_2) < 0.01 * n

def test_paralleltransport():
    '''Test parallel transport of vectors in a simple, analytical example.'''
    dir1 = np.array([[1,-1,0],
                     [1,-1,0],
                     [1,-1,0],
                     [1,-1,0],
                     [1,-1,0]])
    dir2 = np.array([[1,-1,0],
                     [1,1,0],
                     [1,1,0],
                     [1,0.2,0],
                     [1,0.2, 0]])
    dir1 = utils.norm_vector(dir1)
    dir2 = utils.norm_vector(dir2)

    polin = np.array([[1,1,1],
                      [0,0,1],
                      [1, 1,0],
                      [0,0,1],
                      [1, 1,0]])
    polin = utils.norm_vector(polin)

    pmat = paralleltransport_matrix(dir1, dir2)
    out = np.einsum('ijk,ik->ij', pmat, polin)
    # for dir1==dir2 result is unchanged
    assert np.allclose(polin[0, :], out[0, :])
    # s vector is unchanged
    assert np.allclose(polin[:, 2], out[:, 2])
    # result has len 1
    assert np.allclose(np.linalg.norm(out, axis=1), 1)
    # result is perpendicular to dir2
    assert np.allclose(np.einsum('ij,ij->i', out, dir2), 0)

    pmat = paralleltransport_matrix(dir1, dir2, replace_nans=False)
    out = np.einsum('ijk,ik->ij', pmat, polin)
    # dir1==dir2 gives nan
    assert all(np.isnan(out[0, :]))



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

from ..optics import FlatOpticalElement
from ..math.utils import h2e

class ProjectOntoPlane(FlatOpticalElement):
    '''Project photon positions onto a plane.

    *Projection* as used in this class differs from propagating the photons
    until they intersect with a plane. Instead, the current photon positions
    are projected parallel to the normal of the plane.

    Note that the default output columns are called proj_x and _y independent
    of the actual orientation of the plane.

    Examples
    --------
    Project some points onto a plane. If no orientation is given to the
    ``ProjectionOntoPlane`` object, the default is to place it in the yz-plane.

    >>> from astropy.table import Table
    >>> import numpy as np
    >>> from marxs.analysis.coords import ProjectOntoPlane
    >>> photons = Table()
    >>> photons['pos'] = np.array([[3, 3, 3, 1], [12, -1, 0, 1]])
    >>> yzplane = ProjectOntoPlane()
    >>> photons = yzplane(photons)
    >>> photons['proj_x'].data
    array([ 3., -1.])
    >>> photons['proj_y'].data
    array([3., 0.])
    '''

    loc_coos_name = ['proj_x', 'proj_y']
    '''name for output columns of the projected position in plane coordinates.'''

    display = {'shape': 'None'}

    def __call__(self, photons):
        vec_center_inter = - h2e(self.geometry['center']) + h2e(photons['pos'])
        photons[self.loc_coos_name[0]] = np.dot(vec_center_inter, h2e(self.geometry['e_y']))
        photons[self.loc_coos_name[1]] = np.dot(vec_center_inter, h2e(self.geometry['e_z']))
        return photons



from marxs.design.tolerancing import load_and_plot

# In this example, we use a file that's included in MARXS for test purposes
from astropy.utils.data import get_pkg_data_filename
filename = get_pkg_data_filename('data/wiggle_global.fits', 'marxs.design.tests')

tab, fig = load_and_plot(filename, ['dx', 'dy', 'dz', 'rx', 'ry', 'rz'])

fig.axes[0].legend()

for i in range(6):
    fig.axes[2 * i].set_ylim([0, 1e3])
    fig.axes[2 * i + 1].set_ylim([0, 70])

fig.set_size_inches(8, 6)
fig.subplots_adjust(left=.09, right=.93, top=.95, wspace=1.)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from scipy.stats import kstest
from astropy.table import Table, Column
from astropy.coordinates import SkyCoord
import astropy.units as u

from transforms3d.axangles import axangle2aff

import pytest

import marxs
import marxs.optics
import marxs.optics.base
import marxs.optics.aperture
import marxs.source
from marxs.utils import generate_test_photons

@pytest.fixture(autouse=True)
def photons1000():
    '''Make a list of photons parallel to optical axis'''
    mysource = marxs.source.PointSource(coords=SkyCoord(30., 30., unit="deg"))
    np.random.seed(0)
    p = mysource.generate_photons(300 * u.ks)
    mypointing = marxs.source.FixedPointing(coords=SkyCoord(30., 30., unit='deg'))
    return mypointing(p)


def test_pos4d_no_transformation():
    '''Test coordinate transforms on initialization of optical elements.

    This test checks that the appropriate transformations are done.
    It should be extended, once derived properties (e.g. a plane constructed
    from transformed points are implemented.
    '''
    oe = marxs.optics.aperture.RectangleAperture()
    assert np.all(oe.pos4d == np.eye(4))
    assert np.all(oe.geometry['center'] == np.array([0, 0, 0, 1]))
    assert np.all(oe.geometry['e_y'] == np.array([0, 1, 0, 0]))
    assert np.all(oe.geometry['e_z'] == np.array([0, 0, 1, 0]))


def test_pos4d_rotation():
    rotation = axangle2aff(np.array([1, 1, 0]), np.deg2rad(90))
    oe = marxs.optics.aperture.RectangleAperture(orientation=rotation[:3,:3])
    assert np.all(oe.pos4d == rotation)
    assert np.all(oe.geometry['center'] == np.array([0, 0, 0, 1]))
    assert np.allclose(oe.geometry['e_y'], np.array([.5, .5, 1. / np.sqrt(2), 0]))
    assert np.allclose(oe.geometry['e_z'], np.array([1., -1, 0, 0]) / np.sqrt(2))


def test_pos4d_translation():
    oe = marxs.optics.aperture.RectangleAperture(position = np.array([-2, 3, 4.3]))
    expectedpos4d = np.eye(4)
    expectedpos4d[:3, 3] = np.array([-2, 3, 4.3])
    assert np.all(oe.pos4d == expectedpos4d)
    assert np.all(oe.geometry['center'] == np.array([-2, 3, 4.3, 1]))
    assert np.all(oe.geometry['e_y'] == np.array([0, 1, 0, 0]))
    assert np.all(oe.geometry['e_z'] == np.array([0, 0, 1, 0]))



mark = pytest.mark.parametrize

all_slits = [marxs.optics.aperture.RectangleAperture(zoom=2),
             marxs.optics.aperture.RectangleAperture(zoom=[2,2,2]),
            ]

@mark('myslit', all_slits)
def test_pos4d_transforms_slit(photons1000, myslit):
    '''Test coordinate transforms on initialization of optical elements.

    The initial 4D transforms should be done to any optical element.
    Here, I pick the entrance aperture for testing, because it places the
    positional vector of the plucker coordinates in a plane, independent
    of the initial values.
    '''

    p = myslit(photons1000)
    assert np.allclose(p['pos'][:, 0], 0)
    assert kstest((p['pos'][:, 1] + 2) / 4, "uniform")[1] > 0.01
    assert kstest((p['pos'][:, 2] + 2) / 4, "uniform")[1] > 0.01


def test_pos4d_transforms_slit_rotated(photons1000):
    '''Test coordinate transforms on rotated entrance aperture.'''
    p = photons1000

    rotation = axangle2aff(np.array([0, 1, 0]), np.deg2rad(90))
    myslit = marxs.optics.aperture.RectangleAperture(orientation=rotation[:3, :3], zoom=0.5)
    p = myslit(p)
    assert np.allclose(p['pos'][:, 2], 0)
    assert kstest(p['pos'][:, 0] + 0.5, "uniform")[1] > 0.01
    assert kstest(p['pos'][:, 1] + 0.5, "uniform")[1] > 0.01
    # Sometimes fails, because rays exactly parallel to aperture.
    # Find other example, e.g. 45 deg.
    # Check in aperture if that's the case and raise warning.


def test_photonlocalcoords_decorator():

    pos = np.array([[1, 0, 0, 1], [0, 1, 0, 1]])
    dir = np.array([[1, 0, 0, 0], [0, 1, 0, 0]])
    photontab = Table([Column(data=pos, name='pos'),
                       Column(data=dir, name='dir')
                       ])

    class OEforTest(marxs.optics.base.OpticalElement):
        @marxs.optics.base.photonlocalcoords
        def functiontodecorate(self, photons):
            assert np.allclose(photons['dir'], dir)
            assert np.allclose(photons['pos'], np.array([[2., 2, 3, 1], [1, 3, 3, 1]]))
            return photons

    oetest = OEforTest(position=np.array([-1., -2, -3]))
    photontab = oetest.functiontodecorate(photontab)
    assert np.allclose(photontab['dir'], dir)
    assert np.allclose(photontab['pos'], pos)


def test_FlatStack():
    '''Run a stack of two elements and check that both are applied to the photons.'''
    fs = marxs.optics.FlatStack(position=[0, 5, 0], zoom=2,
                                elements=[marxs.optics.EnergyFilter, marxs.optics.FlatDetector],
                                keywords=[{'filterfunc': lambda x: 0.5},{}])
    # Check all layers are at the same position
    assert np.allclose(fs.geometry['center'], np.array([0, 5, 0, 1]))
    assert np.allclose(fs.elements[0].geometry['center'], np.array([0, 5, 0, 1]))
    assert np.allclose(fs.elements[1].geometry['center'], np.array([0, 5, 0, 1]))
    assert np.allclose(fs.pos4d, fs.elements[1].pos4d)

    p = generate_test_photons(5)
    # Photons 0, 1 miss the stack
    p['pos'][:, 1] = [0, 1, 3.1, 4, 5]
    fs.loc_coos_name = ['a', 'b']
    p = fs(p)
    assert np.allclose(p['probability'], [1, 1, .5, .5, .5])
    assert np.all(np.isnan(p['a'][:2]))
    assert np.allclose(p['a'][2:], [-1.9, -1., 0])



import copy
import numpy as np
import matplotlib.pyplot as plt
from marxs import optics, design, analysis, source, simulator
from marxs.design import rowland
from astropy.coordinates import SkyCoord
import astropy.units as u


rowland = design.rowland.RowlandTorus(5000, 5000)
aper = optics.CircleAperture(position=[12000, 0, 0],
                             zoom=[1, 1000, 1000],
                             r_inner=960)
mirr = optics.FlatStack(position=[11000, 0, 0], zoom=[20, 1000, 1000],
                        elements=[optics.PerfectLens,
                                  optics.RadialMirrorScatter],
                        keywords = [{'focallength': 11000},
                                    {'inplanescatter': 4 * u.arcsec,
                                     'perpplanescatter': 0.6 * u.arcsec}])
mirr.display['opacity'] = 0.5

class ColoredFlatGrating(optics.FlatGrating):
    '''Flat grating that also assigns a color to all passing photons.'''
    def specific_process_photons(self, *args, **kwargs):
        out = super(ColoredFlatGrating, self).specific_process_photons(*args, **kwargs)
        out['colorid'] = self.colorid if hasattr(self, 'colorid') else np.nan
        return out

gas = design.rowland.GratingArrayStructure(rowland=rowland, d_element=180,
                                           x_range=[8000, 10000],
                                           radius=[870, 910],
                                           elem_class=ColoredFlatGrating,
                                           elem_args={'d': 2e-4, 'zoom': [1, 80, 80],
                                                      'order_selector': optics.OrderSelector([-1, 0, 1])})
det_kwargs = {'d_element': 105, 'elem_class': optics.FlatDetector,
              'elem_args': {'zoom': [1, 50, 20], 'pixsize': 0.01}}
det = design.rowland.RowlandCircleArray(rowland, theta=[3.1, 3.2], **det_kwargs)
projectfp = analysis.ProjectOntoPlane()

gratingcolors = 'bgr'
for e in gas.elements:
    e.display = copy.deepcopy(e.display)
    e.ang = (np.arctan(e.pos4d[1,3] / e.pos4d[2, 3]) + np.pi/2) % (np.pi)
    e.colorid = int(e.ang / np.pi * 3)
    e.display['color'] = gratingcolors[e.colorid]

instrum = simulator.Sequence(elements=[aper, mirr, gas, det, projectfp])
star = source.PointSource(coords=SkyCoord(30., 30., unit='deg'),
                          energy=1. * u.keV, flux=1. / u.s / u.cm**2)
pointing = source.FixedPointing(coords=SkyCoord(30., 30., unit='deg'))

photons = star.generate_photons(4000 * u.s)
photons = pointing(photons)
photons = instrum(photons)
ind = (photons['probability'] > 0) & (photons['facet'] >=0)

fig = plt.figure()
ax0 = fig.add_subplot(221, aspect='equal')
ax1 = fig.add_subplot(222, aspect='equal', sharey=ax0)
ax0h = fig.add_subplot(223, sharex=ax0)
ax1h = fig.add_subplot(224, sharex=ax1, sharey=ax0h)

ind0 = photons['order'] == 0
# It is arbitrary which direction if called +1 or -1 order
ind1 = np.abs(photons['order']) == 1
out = ax0h.hist(photons['proj_x'][ind & ind0], range=[-1, 1], bins=20,
                lw=0, color='0.8')
out = ax1h.hist(photons['proj_x'][ind & ind1], range=[61, 63], bins=20,
                lw=0, color='0.8')

for i in [0, 2, 1]:  # make green plotted on top
    c = gratingcolors[i]
    indcolor = photons['colorid'] == i
    ax0.plot(photons['proj_x'][ind & ind0 & indcolor],
             photons['proj_y'][ind & ind0 & indcolor],
             '.', color=c)
    ax1.plot(photons['proj_x'][ind & ind1 & indcolor],
             photons['proj_y'][ind & ind1 & indcolor],
             '.', color=c)
    out = ax0h.hist(photons['proj_x'][ind & ind0 & indcolor],
                    range=[-1, 1], bins=20,
                    histtype='step', color=c, lw=2)
    out = ax1h.hist(photons['proj_x'][ind & ind1 & indcolor],
                    range=[61, 63], bins=20,
                    histtype='step', color=c, lw=2)

ax0.set_xlim([-.5,0.5])
ax1.set_xlim([61.65, 62.4])
ax0h.set_xlim([-.5,0.5])
ax1h.set_xlim([61.65, 62.4])

ax0.text(0.1, 0.9, '0th order', transform=ax0.transAxes)
ax1.text(0.1, 0.9, '1st order', transform=ax1.transAxes)
ax0h.set_xlabel('dispersion direction [mm]')
ax0.set_ylabel('cross-dispersion [mm]')
ax0.tick_params(axis='x', labelbottom='off')
ax1.tick_params(axis='both', labelleft='off', labelbottom='off')
ax1h.tick_params(axis='y', labelleft='off')
ax0h.set_ylabel('photons / bin')

for ax in [ax0, ax1, ax0h, ax1h]:
    ax.xaxis.set_major_locator(plt.MaxNLocator(4))
    ax.yaxis.set_major_locator(plt.MaxNLocator(4))

fig.subplots_adjust(hspace=0, wspace=0, left=.18, right=.96, top=.97, bottom=.1)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table
from ..baffle import Baffle


def test_photons_through():
    ''' tests that three photons go through or hit the baffle as predicted'''
    pos = np.array([[1., 0., 0., 1],
                    [1., 0., 0., 1],
                    [1., 0., 0., 1]])
    dir = np.array([[-1., 0., 0., 0],
                    [-1., 1.4, 0.5, 0],
                    [-1., -0.7, 0.2, 0]])
    photons = Table({'pos': pos, 'dir': dir, 'energy': [1, 1, 1], 'polarization': [1, 2, 3], 'probability': [0.5, 0.6, 0.7]})
    baf = Baffle(zoom=np.array([1., 1.5, 0.3]))
    photons = baf(photons)
    expected = np.array([0.5, 0., 0.7])
    assert np.all(photons['probability'] == expected)



# Licensed under GPL version 3 - see LICENSE.rst
from functools import wraps

import numpy as np
from astropy.table import Table, Row

from ..math.geometry import Geometry, FinitePlane
from ..base import SimulationSequenceElement
from ..simulator import BaseContainer

class OpticalElement(SimulationSequenceElement):
    '''Base class for all optical elements in marxs.

    This class cannot be used to instanciate an optical element directly,
    rather it serves as a base class from with other optical elements will be
    derived.

    At the very minumum, any derived class needs to implement `__call__` which
    typically calls `intersect` and either `process_photon` or
    `process_photons`. If the interaction with the photons (e.g. scattering of
    a mirror surface) can be implemented in a vectorized way using numpy array
    operations, the derived class should overwrite `process_photons`
    (`process_photon` is not used in this case).  If no vectorized
    implementation is available, it is sufficient to overwrite
    `process_photon`.  Marxs will call `process_photons`, which (if not
    overwritten) contains a simple for-loop to loop over all photons in the
    array and call `process_photon` on each of them.

    '''

    default_geometry = FinitePlane
    '''If no geometry is specified on initialization, an instance of this class will be used.'''

    # SimulationSequenceElement has display none, but now we have a geometry
    # so we don't need that default here.
    display = {}

    def __init__(self, **kwargs):

        geometry = kwargs.pop('geometry', self.default_geometry)
        if isinstance(geometry, Geometry):
            self.geometry = geometry
        elif issubclass(geometry, Geometry):
            self.geometry = geometry(kwargs)

        super(OpticalElement, self).__init__(**kwargs)

        if not hasattr(self, "loc_coos_name"):
            self.loc_coos_name = self.geometry.loc_coos_name

    @property
    def pos4d(self):
        return self.geometry.pos4d

    def process_photon(self, dir, pos, energy, polarization):
        '''Simulate interaction of optical element with a single photon.

        This is called from the `process_photons` method in a loop over all
        photons. That method also collects the output values and inserts them
        into the photon list. ``process_photon`` can return any number of
        values in additon to the required dir, pos, etc.. Define a class
        attribute ``output_columns`` as a list of strings to determine how into
        which column these numbers should be inserted.

        Parameters
        ----------
        dir : `numpy.ndarray`
            4-d direction vector of ray in homogeneous coordinates
        pos : `numpy.ndarray`
            4-d position of last interaction pf the photons with any optical
            element in homogeneous coordinates. Together with ``dir`` this
            determines the equation of the ray.
        energy : float
            Photon energy in keV.
        polarization : float
            Polarization angle of the photons.

        Returns
        -------
        dir : `numpy.ndarray`
            4-d direction vector of ray in homogeneous coordinates
        pos : `numpy.ndarray`
            4-d position of last interaction pf the photons with any optical
            element in homogeneous coordinates. Together with ``dir`` this
            determines the equation of the ray.
        energy : float
            Photon energy in keV.
        polarization : float
            Polarization angle of the photons.
        probability : float
            Probability that the photon passes this optical element. Set to 0
            if the photon is absorbed, to 1 if it passes and to number between
            0 and 1 to express a probability that the photons passes.
        other : floats
            One number per entry in `output_columns`.

        '''
        raise NotImplementedError
        return dir, pos, energy, polarization, probability, any, other, output, columns

    def __call__(self, photons):
        intersect_out = self.geometry.intersect(photons['dir'].data,
                                                photons['pos'].data)
        return self.process_photons(photons, *intersect_out)

    def process_photons(self, photons, intersect, interpos, intercoos):
        '''Simulate interaction of optical element with photons - vectorized.

        Derived classes should overwrite this function or `process_photon`.

        Parameters
        ----------
        photons: `astropy.table.Table` or `astropy.table.Row`
            Table with photon properties
        intersect : array
            Boolean array marking which photons should be processed by this
            element.
        interpos, intercoos : array (N, 4)
            The array ``interpos`` contains the intersection points in the
            global coordinate system, ``intercoos`` in a local coordiante
            system (2d in most cases).

        Returns
        -------
        photons: `astropy.table.Table` or `astropy.table.Row`
            Table with photon properties.
            If possible, the input table is modified in place, but in some
            cases this might not be possible and the returned Table may be
            a copy. Do not rely on either - use ``photons.copy()`` if you want
            to ensure you are working with an independent copy.

        '''
        if intersect.sum() > 0:
            if hasattr(self, "specific_process_photons"):
                outcols = self.specific_process_photons(photons, intersect, interpos, intercoos)
                self.add_output_cols(photons, list(outcols.keys()))
                for col in outcols:
                    if col == 'probability':
                        photons[col][intersect] *= np.asarray(outcols[col])
                    else:
                        photons[col][intersect] = outcols[col]

            elif hasattr(self, "process_photon"):
                if isinstance(photons, Row):
                    photons = Table(photons)
                outcols = ['dir', 'pos', 'energy', 'polarization', 'probability'] + self.output_columns
                n_intersect = intersect.nonzero()[0]
                for photon, i in zip(photons[intersect], n_intersect):
                    outs = self.process_photon(photon['dir'], photon['pos'],
                                               photon['energy'],
                                               photon['polarization'])
                    for a, b in zip(outcols, outs):
                        if a == 'probability':
                            photons['probability'][i] *= b
                        else:
                            photons[a][i] = b
            else:
                raise AttributeError('Optical element must have one of three: specific_process_photons, process_photon, or override process_photons.')

            self.add_output_cols(photons, self.loc_coos_name)
            # Add ID number to ID col, if requested
            if self.id_col is not None:
                photons[self.id_col][intersect] = self.id_num
            # Set position in different coordinate systems
            photons['pos'][intersect] = interpos[intersect]
            photons[self.loc_coos_name[0]][intersect] = intercoos[intersect, 0]
            photons[self.loc_coos_name[1]][intersect] = intercoos[intersect, 1]

        return photons


class FlatOpticalElement(OpticalElement):
    pass


class FlatStack(FlatOpticalElement, BaseContainer):
    '''Convenience class for several flat, stacked optical elements.

    This class is meant to simplify the specification of a single physical
    element that fullfills several logical functions, e.g. a detector can be
    seen as a a sequence of a contamination layer (which modifies the
    probability of a photon reaching the CCD), a QE filter (which modifies the
    probability of detecting the photon), and the pixelated CCD (which sorts
    the photons in pixels). All these things can be approximated as happening
    in the same physical location, and thus it is convenient to treat all
    three functions as one element.

    Parameters
    ----------
    elements : list of classes
        List of class names specifying the layers in the stack
    keywords : list of dicts
        Dictionaries specifying the properties of each layer (do not set the
        position of individual elements)

    Examples
    --------
    In this example, we will define a single flat CCD with a QE of 0.5 for all
    energies.

    >>> from marxs.optics import FlatStack, EnergyFilter, FlatDetector
    >>> myccd = FlatStack(position=[0, 2, 2], zoom=2,
    ...     elements=[EnergyFilter, FlatDetector],
    ...     keywords=[{'filterfunc': lambda x: 0.5}, {'pixsize': 0.05}])

    '''

    def __init__(self, **kwargs):
        elements = kwargs.pop('elements')
        keywords = kwargs.pop('keywords')
        super(FlatStack, self).__init__(**kwargs)
        self.elements = []
        for elem, k in zip(elements, keywords):
            self.elements.append(elem(pos4d=self.pos4d, **k))

    def specific_process_photons(self, *args, **kwargs):
        return {}

    def process_photons(self, photons, intersect=None, interpos=None, intercoos=None):
        '''
        Parameters
        ----------
        intersect, interpos, intercoos : array (N, 4)
            The array ``interpos`` contains the intersection points in the global
            coordinate system, ``intercoos`` in the local (y,z) system of the grating.
        '''
        if intersect.sum() > 0:
            # This line calls FlatOpticalElement.process_photons to add ID cols
            # and local coos if requested (this could also be done by any of
            # the contained sequence elements, but we want the user to be able
            # to specify that for either of them).
            photons = super(FlatStack, self).process_photons(photons, intersect, interpos, intercoos)
            for e in self.elements:
                photons = e.process_photons(photons, intersect, interpos, intercoos)

        return photons


def photonlocalcoords(f, colnames=['pos', 'dir']):
    '''Decorator for calculation that require a local coordinate system

    This is specifically meant to wrap the :meth:`process_photons` methods of
    any :class:`OpticalElement`; the current implementation expects the call
    signature of :meth:`process_photons`.

    This decorator transforms coordinates from the global system to the local
    system before a function call and back to the global system again after
    the function call.

    Parameters
    ----------
    f : callable with signature ``f(self, photons, *args, **kwargs)``
        The function to be decorated. In the signature, ``photons`` is a
        `astropy.table.Table`.
    colnames : list of strings
        List of all column names in the photon table to be transformed into a
        different coordinate system.
    '''
    @wraps(f)
    def wrapper(self, photons, *args, **kwargs):
        # transform to coordsys if single instrument
        invpos4d = np.linalg.inv(self.pos4d)
        for n in colnames:
            photons[n] = np.einsum('...ij,...j', invpos4d, photons[n])
        photons = f(self, photons, *args, **kwargs)
        # transform back into coordsys of satellite
        for n in colnames:
            photons[n] = np.einsum('...ij,...j', self.pos4d, photons[n])
        return photons

    return wrapper



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import astropy.units as u

from .. import energy2wave

def test_energy2wave():
    '''Check units and formula for energy2wave'''
    energies = np.arange(.1, 10., .1)
    np.allclose(energy2wave / energies,
                (energies * u.keV).to(u.mm, equivalencies=u.spectral()).value
                )



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from numpy.core.umath_tests import inner1d
import transforms3d
import transforms3d.euler
import pytest

from ..geometry import Cylinder, FinitePlane
from ...math.utils import h2e


def test_intersect_plane_forwards():

    pos = np.array([[20., 0., 0.8, 1.],  # start above
                    [0., .0, 0.4, 1.],  # start inside
                    [-5., .1, -.2, 1]]) # start below
    dir = np.array([[-.1, 0., 0., 0],
                    [-1., 0, 0.02, 0.],
                    [-2., -0.02, 0.023, 0.]])

    plane = FinitePlane()
    intersect, interpos, inter_local = plane.intersect(dir, pos)
    assert np.all(intersect == [True, True, False])
    intersect, interpos, inter_local = plane.intersect(-dir, pos)
    assert np.all(intersect == [False, True, True])

def test_intersect_plane_zoomed():
    pos = np.array([[1., 4., 0.8, 1.],
                    [1., -4., 8., 1.],
                    [1., -6., -8., 1]])
    dir = np.array([[-.1, 0., 0., 0],
                    [-1., 0, 0.02, 0.],
                    [-2., -0.02, 0.023, 0.]])

    plane = FinitePlane({'zoom': [1, 5, 10]})
    intersect, interpos, inter_local = plane.intersect(dir, pos)
    assert np.all(intersect == [True, True, False])
    assert np.all(interpos[0, :] == [0, 4, .8, 1])
    assert np.all(inter_local[0, ] == [4, .8])

def test_intersect_plane_moved():

    pos = np.array([[4.1, 5.5, 6.8, 1.],
                    [5., 4.1, 4.99, 1.],
                    [4.1, 3., 6., 1]])
    dir = np.array([[-.1, 0., 0., 0],
                    [-1., 0, 0.02, 0.],
                    [-2., -0.02, 0.023, 0.]])

    plane = FinitePlane({'position': [4,5,6]})
    intersect, interpos, inter_local = plane.intersect(dir, pos)
    assert np.all(intersect == [True, True, False])
    assert np.all(interpos[1, :] == [4, 4.1, 5.01, 1])
    assert np.allclose(inter_local[1, :], [-0.9, -.99])

def test_intersect_tube_miss():
    '''Rays passing at larger radius or through the center miss the tube.'''
    circ = Cylinder()
    # Passing at larger radius
    intersect, interpos, inter_local = circ.intersect(np.array([[1., 0., .1, 0],
                                                                [-1., -1., 0., 0.]]),
                                                         np.array([[0, -1.1, 0., 1.],
                                                                   [2., 0., 0., 1.]]))
    assert np.all(intersect == False)
    assert np.all(np.isnan(interpos))
    assert np.all(np.isnan(inter_local))

    # Through the center almost parallel to z-axis
    intersect, interpos, inter_local = circ.intersect(np.array([[0., 0.1, 1., 0]]),
                                                         np.array([[0.5, 0.5, 0., 1.]]))
    assert np.all(intersect == False)
    assert np.all(np.isnan(interpos))
    assert np.all(np.isnan(inter_local))

    # Repeat with a tube that's moved to make sure we did not mess up local and
    # global coordinates in the implementation
    circ = Cylinder({'position': [0.8, 0.8, 1.2]})
    intersect, interpos, inter_local = circ.intersect(np.array([[1., 0., .0, 0],
                                                                [1., 0., 0., 0.]]),
                                                         np.array([[0., 0., 0., 1.],
                                                                   [0., -0.3, 1., 1.]]))
    assert np.all(intersect == False)
    assert np.all(np.isnan(interpos))
    assert np.all(np.isnan(inter_local))

def test_intersect_tube_rays_forward():
    '''Rays only intersect if propagating them forward works'''
    pos = np.array([[20., 0., 0.8, 1.],  # start above
                    [0., .0, 0.4, 1.],  # start inside
                    [-5., .1, -.2, 1]]) # start below
    dir = np.array([[-.1, 0., 0., 0],
                    [-1., 0, 0.02, 0.],
                    [-2., -0.02, 0.023, 0.]])

    circ = Cylinder({'zoom': [1, 1, 1]})
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == [True, True, False])
    intersect, interpos, inter_local = circ.intersect(-dir, pos)
    assert np.all(intersect == [False, True, True])


def test_intersect_tube_hitmiss_zoomed():
    '''Rays that hit one tube, but miss a shorter one'''
    dir = np.array([[-.1, 0., 0., 0], [-1., 0, 0.2, 0.]])
    pos = np.array([[20., 0., 0.8, 1.], [2., .0, 0.4, 1.]])

    circ = Cylinder({'zoom': [1, 1, 1]})
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)

    circ = Cylinder({'zoom': [1, 1, .5]})
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == False)
    assert np.all(np.isnan(interpos))
    assert np.all(np.isnan(inter_local))

def test_intersect_tube_2points():
    dir = np.array([[-.1, 0., 0., 0], [-0.5, 0, 0., 0.], [-1., 0., 0., 0.]])
    pos = np.array([[50., 0., 0., 1.], [10., .5, 0., 1.], [2, 0, .3, 1.]])

    circ = Cylinder()
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), np.array([[1., 0., 0.],
                                                [np.sqrt(0.75), 0.5, 0.],
                                                [1, 0., .3]]))
    assert np.allclose(inter_local, np.array([[0., np.arcsin(0.5), 0.],
                                              [0., 0., 0.3]]).T)

def test_intersect_tube_2points_transformed():
    '''inter_local should be the same as interpos,
    even if we make the tube longer in z'''
    pos = np.array([[50., 0., 0., 1.], [10., .5, 0., 1.], [2, 0, .3, 1.]])
    dir = np.array([[-.1, 0., 0., 0], [-0.5, 0, 0., 0.], [-1., 0., 0., 0.]])

    exp_inter_loc = np.array([[0., np.arcsin(0.5), 0.],
                              [0., 0., 0.3]]).T
    exp_interpos = np.array([[1., 0., 0.],
                             [np.sqrt(0.75), 0.5, 0.],
                             [1, 0., .3]])
    circ = Cylinder({'zoom': [1, 1., 5.]})
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), exp_interpos)
    assert np.allclose(inter_local, exp_inter_loc)
    # Rotate: interpos should be the same as in the unrotated case.
    # Beware of rotation direction!
    orient = transforms3d.axangles.axangle2mat([0, 0, 1], -.3)
    circ = Cylinder({'orientation': orient})
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), exp_interpos)
    assert np.allclose(inter_local, np.array([[0.3, 0.3 + np.arcsin(0.5), 0.3],
                                              [0., 0., 0.3]]).T)

    # Shift: inter_local is the same, but interpos changes
    circ = Cylinder({'position': [-.3, 0, 0]})
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), np.array([[.7, 0, 0],
                                                [np.sqrt(0.75) -.3, 0.5, 0.],
                                                [.7, 0., .3]]))
    assert np.allclose(inter_local, exp_inter_loc)

def test_intersect_tube_2points_outside():
    '''Setting phi_lim

    These are the same numbers, but with the phi_lim set, all intersection are
    at the "bottom" of the cylinder.'''
    pos = np.array([[50., 0., 0., 1.], [10., .5, 0., 1.], [2, 0, .3, 1.]])
    dir = np.array([[-.1, 0., 0., 0], [-0.5, 0, 0., 0.], [-1., 0., 0., 0.]])

    circ = Cylinder({'phi_lim': [.6, np.pi]})
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), np.array([[-1., 0., 0.],
                                                [-np.sqrt(0.75), 0.5, 0.],
                                                [-1, 0., .3]]))
    assert np.allclose(inter_local, np.array([[np.pi, np.pi - np.arcsin(0.5), np.pi],
                                              [0., 0., 0.3]]).T)
    # phi_lims that guarantee a miss
    circ = Cylinder({'phi_lim': [-.4, -.3]})
    intersect, interpos, inter_local = circ.intersect(dir, pos)
    assert np.all(intersect == False)
    assert np.all(np.isnan(interpos))
    assert np.all(np.isnan(inter_local))

@pytest.mark.xfail
def test_phi_lim_verification():
    '''Check Errors for wrong phi_lim format.'''
    with pytest.raises(ValueError) as e:
        circ = Cylinder({'phi_lim': [-.4, -.3, .4]})
    assert '[lower limit, upper limit]' in str(e.value)

    with pytest.raises(ValueError) as e:
        circ = Cylinder({'phi_lim': [-.4, -.5]})
    assert '[lower limit, upper limit]' in str(e.value)

    with pytest.raises(ValueError) as e:
        circ = Cylinder({'phi_lim': [-.4, 5]})
    assert 'range -pi to +pi' in str(e.value)


def test_intersect_tube_2points_translation():
    '''Repeat test above with a tube that's moved and zoomed to make sure we
    did not mess up local and global coordinates in the implementation
    '''
    circ = Cylinder({'position': [0.8, 0.8, 1.2], 'zoom': [1, 1, 2]})
    intersect, interpos, inter_local = circ.intersect(np.array([[1., 0., .0, 0],
                                                                [1., 0., 0., 0.]]),
                                                         np.array([[0., 0., 0., 1.],
                                                                   [0., 0., 1., 1.]]))
    assert np.all(intersect == True)
    assert np.allclose(h2e(interpos), np.array([[.2, 0., 0.],
                                                [.2, 0., 1.]]))

def test_tube_parametric():
    '''Generate points on surface using parametric. Then make rays for intersection
    which should return those points as intersection points.'''
    circ = Cylinder({'zoom': [2., 3., 4.]})
    parametric = circ.parametric_surface(phi=[-.1, 0., 1.])
    parametric = parametric.reshape((6, 4))
    # select dir so that it's in the right direction to recover these points.
    dir = np.tile([1, 0, 0, 0], (6, 1))
    intersect, interpos, inter_local = circ.intersect(dir, parametric)
    assert np.allclose(parametric, interpos)

@pytest.mark.parametrize("geom", [FinitePlane, Cylinder])
def test_local_coordsys(geom):
    '''Ensure local coordiante systems are orthonormal'''
    rot = transforms3d.euler.euler2mat(*(np.pi * 2 * np.random.rand(3)))
    g = geom({'rotation': rot,
              'position': np.random.rand(3)})

    x, y, z = g.get_local_euklid_bases(np.random.rand(5, 2))

    assert np.allclose(inner1d(x, y), 0)
    assert np.allclose(inner1d(x, z), 0)

    for vec in [x, y, z]:
        assert np.allclose(np.linalg.norm(vec, axis=1), 1.)

    # Check it's a right-handed coordinage system
    assert np.allclose(np.cross(x[:, :3], y[:, :3]), z[:, :3])



import numpy as np
from matplotlib import pyplot as plt
from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.table import QTable

from marxs.source import poisson_process
from marxs.missions import chandra
from marxs.source import PointSource

ngc1313_X1 = SkyCoord("3h18m19.99s -66d29m10.97s")

energies = np.arange(.3, 8., .01) * u.keV
fluxdensity = 6.3e-4 * energies.value**(-1.9) / u.s / u.cm**2 / u.keV
fluxperbin = fluxdensity[1:] * np.diff(energies)
flux = poisson_process(fluxperbin.sum())
energytab = QTable({'energy': energies, 'fluxdensity': fluxdensity})

aperture = chandra.Aperture()

src1 = PointSource(coords=ngc1313_X1, energy=energytab,
                   flux=flux, geomarea=aperture.area)

pointing = chandra.LissajousDither(coords=ngc1313_X1)
hrma = chandra.HRMA()
acis = chandra.ACIS(chips=[4,5,6,7,8,9], aimpoint=chandra.AIMPOINTS['ACIS-S'])

photons = src1.generate_photons(5 * u.ks)
photons = pointing(photons)
photons = aperture(photons)
photons = hrma(photons)
photons = acis(photons)

line = plt.plot(photons['tdetx'], photons['tdety'], '.')



# Licensed under GPL version 3 - see LICENSE.rst
from .analysis import (find_best_detector_position, detected_fraction)
from .gratings import (resolvingpower_per_order,
                       resolvingpower_from_photonlist,
                       weighted_per_order)
from .coords import ProjectOntoPlane



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import astropy.units as u
from numpy.core.umath_tests import inner1d

from .utils import norm_vector, e2h


__all__ = ['polarization_vectors', 'Q_reflection', 'paralleltransport_matrix',
           'parallel_transport']


def polarization_vectors(dir_array, angles):
    '''Converts polarization angles to vectors in the direction of polarization.

    Follows convention: Vector perpendicular to photon direction and closest to +y axis is
    angle 0 for polarization direction, unless photon direction is parallel to the y axis,
    in which case the vector closest to the +x axis is angle 0.

    Parameters
    ----------
    dir_array : nx4 np.array
        each row is the homogeneous coordinates for a photon's direction vector
    angles : np.array
        1D array with the polarization angles
    '''
    n = len(angles)
    polarization = np.zeros((n, 4))
    x = np.array([1., 0., 0.])
    y = np.array([0., 1., 0.])

    #	NOTE: The commented code works and is more readable, but the current code is faster.
    #	for i in range(0, n):
    #		r = h2e(dir_array[i])
    #		r /= np.linalg.norm(r)
    #		if not (np.isclose(r[0], 0.) and np.isclose(r[2], 0.)):
    #			# polarization relative to positive y at 0
    #			v_1 = y - (r * np.dot(r, y))
    #			v_1 /= np.linalg.norm(v_1)
    #		else:
    #			# polarization relative to positive x at 0
    #			v_1 = x - (r * np.dot(r, x))
    #			v_1 /= np.linalg.norm(v_1)
    #
    #		# right hand coordinate system is v_1, v_2, r (photon direction)
    #		v_2 = np.cross(r, v_1)
    #		polarization[i, 0:3] = v_1 * np.cos(angles[i]) + v_2 * np.sin(angles[i])
    #		polarization[i, 3] = 0

    r = dir_array.copy()[:,0:3]
    r /= np.linalg.norm(r, axis=1)[:, np.newaxis]
    pol_convention_x = np.isclose(r[:,0], 0.) & np.isclose(r[:,2], 0.)
    if hasattr(angles, "unit") and (angles.unit is not None):
        angles = angles.to(u.rad)
    # polarization relative to positive y or x at 0
    v_1 = ~pol_convention_x[:, np.newaxis] * (y - r * np.dot(r, y)[:, np.newaxis])
    v_1 += pol_convention_x[:, np.newaxis] * (x - r * np.dot(r, x)[:, np.newaxis])
    v_1 /= np.linalg.norm(v_1, axis=1)[:, np.newaxis]

    # right hand coordinate system is v_1, v_2, r (photon direction)
    v_2 = np.cross(r, v_1)
    polarization[:, 0:3] = v_1 * np.cos(angles)[:, np.newaxis] + v_2 * np.sin(angles)[:, np.newaxis]

    return polarization

def Q_reflection(delta_dir):
    '''Reflection of a polarization vector on a non-polarizing surface.

    This can also be used for other elements that change the direction of the
    photon without adding any more polarization and where both sides
    propagate in the same medium.
    See `Yun (2011) <http://hdl.handle.net/10150/202979>`_, eqn 4.3.13 for details.

    Parameters
    ----------
    delta_dir : np.array of shape (n, 4)
        Array of photon direction coordinates in homogeneous coordinates:
        ``delta_dir = photon['dir_old'] - photons['dir_new']``.
        Note that this vector is **not** normalized.

    Returns
    -------
    q : np.array of shape (n, 4, 4)
        Array of parallel transport ray tracing matrices.
    '''
    if delta_dir.shape != 2:
        raise ValueError('delta_dir must have dimension (n, 4).')
    m = delta_dir[..., None, :] * delta_dir[..., :, None]
    return np.eye(4) - 2 / (np.linalg.norm(delta_dir, axis=1)**2)[:, None, None] * m

def paralleltransport_matrix(dir1, dir2, jones=np.eye(2), replace_nans=True):
    '''Calculate parallel transport ray tracing matrix.

    Parallel transport for a vector implies that the component s
    (perpendicular, from German *senkrecht*) to the planes spanned by
    ``dir1`` and ``dir2`` stays the same. If ``dir1`` is parallel to ``dir2``
    this plane is not well defined and the resulting matrix elements will
    be set to ``np.nan``, unless ``replace_nans`` is set.

    Note that the ray matrix returned works on an eukledian 3d vector, not a
    homogeneous vector. (Polarization is a vector, thus the forth element of the
    homogeneous vector is always 0 and returning (4,4) matrices is just a waste
    of space.)

    Parameters
    ----------
    dir1, dir2 : np.array of shape (n, 3)
        Direction before and after the interaction.
    jones : np.array of shape (2,2)
        Jones matrix in the local s,p system of the optical element.
    replace_nans : bool
        If ``True`` return an identity matrix for those rays with
        ``dir1=dir2``. In those cases, the local coordinate system is not well
        defined and thus no Jones matrix can be applied. In MARXS ``dir1=dir2``
        often happens if some photons in a list miss the optical element in
        question - these photons just pass through and their polarization vector
        should be unchanged.

    Returns
    -------
    p_mat : np.array of shape(n, 3, 3)
    '''
    dir1 = norm_vector(dir1)
    dir2 = norm_vector(dir2)

    jones_3 = np.eye(3)
    jones_3[:2, :2] = jones

    pmat = np.zeros((dir1.shape[0], 3, 3))
    s = np.cross(dir1, dir2)
    s_norm = np.linalg.norm(s, axis=1)
    # Find dir values that remain unchanged
    # For these the cross prodict will by 0
    # and a numerical error is raised in s / norm(s)
    # Expected output value for these depends on "replace_nans"
    ind = np.isclose(s_norm, 0)

    if (~ind).sum() > 0:

        s = s[~ind, :] / s_norm[~ind][:, None]
        p_in = np.cross(dir1[~ind, :], s)
        p_out = np.cross(dir2[~ind, :], s)
        Oininv = np.array([s, p_in, dir1[~ind, :]]).swapaxes(1, 0)
        Oout = np.array([s, p_out, dir2[~ind, :]]).swapaxes(1, 2).T
        temp = np.einsum('...ij,kjl->kil', jones_3, Oininv)
        pmat[~ind, :, :] = np.einsum('ijk,ikl->ijl', Oout, temp)

    factor = 1 if replace_nans else np.nan
    pmat[ind, :, :] = factor * np.eye(3)[None, :, :]
    return pmat

def parallel_transport(dir_old, dir_new, pol_old, **kwargs):
    '''Parallel transport of the polarization vector with no polarization happening.

    Parameters
    ----------
    dir_old, dir_new : np.array of shape (n, 4)
        Old and new photon direction in homogeneous coordinates.
    pol_old : np.array of shape (n, 4)
        Old polarization vector in homogeneous coordinates.
    kwargs : dict
        All other arguments are passed on to `~marxs.math.polarization.paralleltransport_matrix`.
    Returns
    -------
    pol : np.array of shape (m, 4)
        Parallel transported vectors.
    '''
    pmat = paralleltransport_matrix(dir_old[:, :3], dir_new[:, :3])
    out = np.einsum('ijk,ik->ij', pmat, pol_old[:, :3])
    return e2h(out, 0)
